"""
Defines the unit tests for the
:mod:`colour.models.rgb.transfer_functions.arib_std_b67` module.
"""

import numpy as np
import unittest

from colour.models.rgb.transfer_functions import (
    oetf_ARIBSTDB67,
    oetf_inverse_ARIBSTDB67,
)
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestOetf_ARIBSTDB67",
    "TestOetf_inverse_ARIBSTDB67",
]


class TestOetf_ARIBSTDB67(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.arib_std_b67.\
oetf_ARIBSTDB67` definition unit tests methods.
    """

    def test_oetf_ARIBSTDB67(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.arib_std_b67.\
oetf_ARIBSTDB67` definition.
        """

        self.assertAlmostEqual(oetf_ARIBSTDB67(-0.25), -0.25, places=7)

        self.assertAlmostEqual(oetf_ARIBSTDB67(0.0), 0.0, places=7)

        self.assertAlmostEqual(
            oetf_ARIBSTDB67(0.18), 0.212132034355964, places=7
        )

        self.assertAlmostEqual(oetf_ARIBSTDB67(1.0), 0.5, places=7)

        self.assertAlmostEqual(
            oetf_ARIBSTDB67(64.0), 1.302858098046995, places=7
        )

    def test_n_dimensional_oetf_ARIBSTDB67(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.arib_std_b67.\
oetf_ARIBSTDB67` definition n-dimensional arrays support.
        """

        E = 0.18
        E_p = oetf_ARIBSTDB67(E)

        E = np.tile(E, 6)
        E_p = np.tile(E_p, 6)
        np.testing.assert_almost_equal(oetf_ARIBSTDB67(E), E_p, decimal=7)

        E = np.reshape(E, (2, 3))
        E_p = np.reshape(E_p, (2, 3))
        np.testing.assert_almost_equal(oetf_ARIBSTDB67(E), E_p, decimal=7)

        E = np.reshape(E, (2, 3, 1))
        E_p = np.reshape(E_p, (2, 3, 1))
        np.testing.assert_almost_equal(oetf_ARIBSTDB67(E), E_p, decimal=7)

    def test_domain_range_scale_oetf_ARIBSTDB67(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.arib_std_b67.\
oetf_ARIBSTDB67` definition domain and range scale support.
        """

        E = 0.18
        E_p = oetf_ARIBSTDB67(E)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    oetf_ARIBSTDB67(E * factor), E_p * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_oetf_ARIBSTDB67(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.arib_std_b67.\
oetf_ARIBSTDB67` definition nan support.
        """

        oetf_ARIBSTDB67(np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]))


class TestOetf_inverse_ARIBSTDB67(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.arib_std_b67.\
oetf_inverse_ARIBSTDB67` definition unit tests methods.
    """

    def test_oetf_inverse_ARIBSTDB67(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.arib_std_b67.\
oetf_inverse_ARIBSTDB67` definition.
        """

        self.assertAlmostEqual(oetf_inverse_ARIBSTDB67(-0.25), -0.25, places=7)

        self.assertAlmostEqual(oetf_inverse_ARIBSTDB67(0.0), 0.0, places=7)

        self.assertAlmostEqual(
            oetf_inverse_ARIBSTDB67(0.212132034355964), 0.18, places=7
        )

        self.assertAlmostEqual(oetf_inverse_ARIBSTDB67(0.5), 1.0, places=7)

        self.assertAlmostEqual(
            oetf_inverse_ARIBSTDB67(1.302858098046995), 64.0, places=7
        )

    def test_n_dimensional_oetf_inverse_ARIBSTDB67(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.arib_std_b67.\
oetf_inverse_ARIBSTDB67` definition n-dimensional arrays support.
        """

        E_p = 0.212132034355964
        E = oetf_inverse_ARIBSTDB67(E_p)

        E_p = np.tile(E_p, 6)
        E = np.tile(E, 6)
        np.testing.assert_almost_equal(
            oetf_inverse_ARIBSTDB67(E_p), E, decimal=7
        )

        E_p = np.reshape(E_p, (2, 3))
        E = np.reshape(E, (2, 3))
        np.testing.assert_almost_equal(
            oetf_inverse_ARIBSTDB67(E_p), E, decimal=7
        )

        E_p = np.reshape(E_p, (2, 3, 1))
        E = np.reshape(E, (2, 3, 1))
        np.testing.assert_almost_equal(
            oetf_inverse_ARIBSTDB67(E_p), E, decimal=7
        )

    def test_domain_range_scale_oetf_inverse_ARIBSTDB67(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.arib_std_b67.\
oetf_inverse_ARIBSTDB67` definition domain and range scale support.
        """

        E_p = 0.212132034355964
        E = oetf_inverse_ARIBSTDB67(E_p)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    oetf_inverse_ARIBSTDB67(E_p * factor),
                    E * factor,
                    decimal=7,
                )

    @ignore_numpy_errors
    def test_nan_oetf_inverse_ARIBSTDB67(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.arib_std_b67.\
oetf_inverse_ARIBSTDB67` definition nan support.
        """

        oetf_inverse_ARIBSTDB67(
            np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan])
        )


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.blindness.machado2009` module."""

import numpy as np
import unittest

from colour.blindness import (
    CVD_MATRICES_MACHADO2010,
    matrix_cvd_Machado2009,
    msds_cmfs_anomalous_trichromacy_Machado2009,
    matrix_anomalous_trichromacy_Machado2009,
)
from colour.characterisation import MSDS_DISPLAY_PRIMARIES
from colour.colorimetry import MSDS_CMFS_LMS
from colour.utilities import ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestMsdsCmfsAnomalousTrichromacyMachado2009",
    "TestMatrixAnomalousTrichromacyMachado2009",
    "TestMatrixCvdMachado2009",
]


class TestMsdsCmfsAnomalousTrichromacyMachado2009(unittest.TestCase):
    """
    Define :func:`colour.blindness.machado2009.\
msds_cmfs_anomalous_trichromacy_Machado2009` definition unit tests methods.
    """

    def test_msds_cmfs_anomalous_trichromacy_Machado2009(self):
        """
        Test :func:`colour.blindness.machado2009.\
msds_cmfs_anomalous_trichromacy_Machado2009` definition.
        """

        cmfs = MSDS_CMFS_LMS.get("Smith & Pokorny 1975 Normal Trichromats")
        np.testing.assert_almost_equal(
            msds_cmfs_anomalous_trichromacy_Machado2009(
                cmfs,
                np.array(
                    [0, 0, 0],
                ),
            )[450],
            cmfs[450],
            decimal=7,
        )

        np.testing.assert_almost_equal(
            msds_cmfs_anomalous_trichromacy_Machado2009(
                cmfs,
                np.array(
                    [1, 0, 0],
                ),
            )[450],
            np.array([0.03631700, 0.06350000, 0.91000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            msds_cmfs_anomalous_trichromacy_Machado2009(
                cmfs,
                np.array(
                    [0, 1, 0],
                ),
            )[450],
            np.array([0.03430000, 0.06178404, 0.91000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            msds_cmfs_anomalous_trichromacy_Machado2009(
                cmfs,
                np.array(
                    [0, 0, 1],
                ),
            )[450],
            np.array([0.03430000, 0.06350000, 0.92270240]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            msds_cmfs_anomalous_trichromacy_Machado2009(
                cmfs,
                np.array(
                    [10, 0, 0],
                ),
            )[450],
            np.array([0.05447001, 0.06350000, 0.91000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            msds_cmfs_anomalous_trichromacy_Machado2009(
                cmfs,
                np.array(
                    [0, 10, 0],
                ),
            )[450],
            np.array([0.03430000, 0.04634036, 0.91000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            msds_cmfs_anomalous_trichromacy_Machado2009(
                cmfs,
                np.array(
                    [0, 0, 10],
                ),
            )[450],
            np.array([0.03430000, 0.06350000, 1.00000000]),
            decimal=7,
        )


class TestMatrixAnomalousTrichromacyMachado2009(unittest.TestCase):
    """
    Define :func:`colour.blindness.machado2009.\
matrix_anomalous_trichromacy_Machado2009` definition unit tests methods.
    """

    def test_matrix_anomalous_trichromacy_Machado2009(self):
        """
        Test :func:`colour.blindness.machado2009.\
matrix_anomalous_trichromacy_Machado2009` definition.
        """

        cmfs = MSDS_CMFS_LMS.get("Smith & Pokorny 1975 Normal Trichromats")
        primaries = MSDS_DISPLAY_PRIMARIES["Typical CRT Brainard 1997"]
        np.testing.assert_almost_equal(
            matrix_anomalous_trichromacy_Machado2009(
                cmfs, primaries, np.array([0, 0, 0])
            ),
            np.identity(3),
            decimal=7,
        )

        np.testing.assert_allclose(
            matrix_anomalous_trichromacy_Machado2009(
                cmfs, primaries, np.array([2, 0, 0])
            ),
            CVD_MATRICES_MACHADO2010.get("Protanomaly").get(0.1),
            rtol=0.0001,
            atol=0.0001,
        )

        np.testing.assert_allclose(
            matrix_anomalous_trichromacy_Machado2009(
                cmfs, primaries, np.array([20, 0, 0])
            ),
            CVD_MATRICES_MACHADO2010.get("Protanomaly").get(1.0),
            rtol=0.0001,
            atol=0.0001,
        )

        np.testing.assert_allclose(
            matrix_anomalous_trichromacy_Machado2009(
                cmfs, primaries, np.array([0, 2, 0])
            ),
            CVD_MATRICES_MACHADO2010.get("Deuteranomaly").get(0.1),
            rtol=0.0001,
            atol=0.0001,
        )

        np.testing.assert_allclose(
            matrix_anomalous_trichromacy_Machado2009(
                cmfs, primaries, np.array([0, 20, 0])
            ),
            CVD_MATRICES_MACHADO2010.get("Deuteranomaly").get(1.0),
            rtol=0.0001,
            atol=0.0001,
        )

        np.testing.assert_allclose(
            matrix_anomalous_trichromacy_Machado2009(
                cmfs, primaries, np.array([0, 0, 5.00056688094503])
            ),
            CVD_MATRICES_MACHADO2010.get("Tritanomaly").get(0.1),
            rtol=0.0001,
            atol=0.0001,
        )

        np.testing.assert_allclose(
            matrix_anomalous_trichromacy_Machado2009(
                cmfs, primaries, np.array([0, 0, 59.00590434857581])
            ),
            CVD_MATRICES_MACHADO2010.get("Tritanomaly").get(1.0),
            rtol=0.001,
            atol=0.001,
        )


class TestMatrixCvdMachado2009(unittest.TestCase):
    """
    Define :func:`colour.blindness.machado2009.matrix_cvd_Machado2009`
    definition unit tests methods.
    """

    def test_matrix_cvd_Machado2009(self):
        """
        Test :func:`colour.blindness.machado2009.matrix_cvd_Machado2009`
        definition.
        """

        np.testing.assert_almost_equal(
            matrix_cvd_Machado2009("Protanomaly", 0.0),
            np.array(
                [
                    [1, 0, 0],
                    [0, 1, 0],
                    [0, 0, 1],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            matrix_cvd_Machado2009("Deuteranomaly", 0.1),
            np.array(
                [
                    [0.86643500, 0.17770400, -0.04413900],
                    [0.04956700, 0.93906300, 0.01137000],
                    [-0.00345300, 0.00723300, 0.99622000],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            matrix_cvd_Machado2009("Tritanomaly", 1.0),
            np.array(
                [
                    [1.25552800, -0.07674900, -0.17877900],
                    [-0.07841100, 0.93080900, 0.14760200],
                    [0.00473300, 0.69136700, 0.30390000],
                ]
            ),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            matrix_cvd_Machado2009("Tritanomaly", 0.55),
            np.array(
                [
                    [1.06088700, -0.01504350, -0.04584350],
                    [-0.01895750, 0.96774750, 0.05121150],
                    [0.00317700, 0.27513700, 0.72168600],
                ]
            ),
            decimal=7,
        )

    @ignore_numpy_errors
    def test_nan_matrix_cvd_Machado2009(self):
        """
        Test :func:`colour.blindness.machado2009.matrix_cvd_Machado2009`
        definition nan support.
        """

        for case in [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]:
            matrix_cvd_Machado2009("Tritanomaly", case)


if __name__ == "__main__":
    unittest.main()



"""
Signal
======

Defines the class implementing support for continuous signal:

-   :class:`colour.continuous.Signal`
"""

from __future__ import annotations

import numpy as np
from operator import (
    add,
    mul,
    pow,
    sub,
    truediv,
    iadd,
    imul,
    ipow,
    isub,
    itruediv,
)
from collections.abc import Iterator, Mapping, Sequence, ValuesView
from colour.algebra import Extrapolator, KernelInterpolator
from colour.constants import DEFAULT_FLOAT_DTYPE
from colour.continuous import AbstractContinuousFunction
from colour.hints import (
    Any,
    ArrayLike,
    Boolean,
    Callable,
    DTypeFloating,
    Dict,
    FloatingOrArrayLike,
    FloatingOrNDArray,
    Integer,
    Literal,
    NDArray,
    Number,
    Optional,
    Tuple,
    Type,
    TypeExtrapolator,
    TypeInterpolator,
    Union,
    cast,
)
from colour.utilities import (
    as_float_array,
    attest,
    fill_nan,
    full,
    is_pandas_installed,
    optional,
    required,
    runtime_warning,
    tsplit,
    tstack,
    validate_method,
)
from colour.utilities.documentation import is_documentation_building

if is_pandas_installed():
    from pandas import Series
else:  # pragma: no cover
    from unittest import mock

    Series = mock.MagicMock()

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "Signal",
]


class Signal(AbstractContinuousFunction):
    """
    Define the base class for continuous signal.

    The class implements the :meth:`Signal.function` method so that evaluating
    the function for any independent domain variable :math:`x \\in\\mathbb{R}`
    returns a corresponding range variable :math:`y \\in\\mathbb{R}`.
    It adopts an interpolating function encapsulated inside an extrapolating
    function. The resulting function independent domain, stored as discrete
    values in the :attr:`colour.continuous.Signal.domain` property corresponds
    with the function dependent and already known range stored in the
    :attr:`colour.continuous.Signal.range` property.

    .. important::

        Specific documentation about getting, setting, indexing and slicing the
        continuous signal values is available in the
        :ref:`spectral-representation-and-continuous-signal` section.

    Parameters
    ----------
    data
        Data to be stored in the continuous signal.
    domain
        Values to initialise the :attr:`colour.continuous.Signal.domain`
        attribute with. If both ``data`` and ``domain`` arguments are defined,
        the latter with be used to initialise the
        :attr:`colour.continuous.Signal.domain` property.

    Other Parameters
    ----------------
    dtype
        Floating point data type.
    extrapolator
        Extrapolator class type to use as extrapolating function.
    extrapolator_kwargs
        Arguments to use when instantiating the extrapolating function.
    interpolator
        Interpolator class type to use as interpolating function.
    interpolator_kwargs
        Arguments to use when instantiating the interpolating function.
    name
        Continuous signal name.

    Attributes
    ----------
    -   :attr:`~colour.continuous.Signal.dtype`
    -   :attr:`~colour.continuous.Signal.domain`
    -   :attr:`~colour.continuous.Signal.range`
    -   :attr:`~colour.continuous.Signal.interpolator`
    -   :attr:`~colour.continuous.Signal.interpolator_kwargs`
    -   :attr:`~colour.continuous.Signal.extrapolator`
    -   :attr:`~colour.continuous.Signal.extrapolator_kwargs`
    -   :attr:`~colour.continuous.Signal.function`

    Methods
    -------
    -   :meth:`~colour.continuous.Signal.__init__`
    -   :meth:`~colour.continuous.Signal.__str__`
    -   :meth:`~colour.continuous.Signal.__repr__`
    -   :meth:`~colour.continuous.Signal.__hash__`
    -   :meth:`~colour.continuous.Signal.__getitem__`
    -   :meth:`~colour.continuous.Signal.__setitem__`
    -   :meth:`~colour.continuous.Signal.__contains__`
    -   :meth:`~colour.continuous.Signal.__eq__`
    -   :meth:`~colour.continuous.Signal.__ne__`
    -   :meth:`~colour.continuous.Signal.arithmetical_operation`
    -   :meth:`~colour.continuous.Signal.signal_unpack_data`
    -   :meth:`~colour.continuous.Signal.fill_nan`
    -   :meth:`~colour.continuous.Signal.to_series`

    Examples
    --------
    Instantiation with implicit *domain*:

    >>> range_ = np.linspace(10, 100, 10)
    >>> print(Signal(range_))
    [[   0.   10.]
     [   1.   20.]
     [   2.   30.]
     [   3.   40.]
     [   4.   50.]
     [   5.   60.]
     [   6.   70.]
     [   7.   80.]
     [   8.   90.]
     [   9.  100.]]

    Instantiation with explicit *domain*:

    >>> domain = np.arange(100, 1100, 100)
    >>> print(Signal(range_, domain))
    [[  100.    10.]
     [  200.    20.]
     [  300.    30.]
     [  400.    40.]
     [  500.    50.]
     [  600.    60.]
     [  700.    70.]
     [  800.    80.]
     [  900.    90.]
     [ 1000.   100.]]

    Instantiation with a *dict*:

    >>> print(Signal(dict(zip(domain, range_))))
    [[  100.    10.]
     [  200.    20.]
     [  300.    30.]
     [  400.    40.]
     [  500.    50.]
     [  600.    60.]
     [  700.    70.]
     [  800.    80.]
     [  900.    90.]
     [ 1000.   100.]]

    Instantiation with a *Pandas* :class:`pandas.Series`:

    >>> if is_pandas_installed():
    ...     from pandas import Series
    ...     print(Signal(  # doctest: +SKIP
    ...         Series(dict(zip(domain, range_)))))
    [[  100.    10.]
     [  200.    20.]
     [  300.    30.]
     [  400.    40.]
     [  500.    50.]
     [  600.    60.]
     [  700.    70.]
     [  800.    80.]
     [  900.    90.]
     [ 1000.   100.]]

    Retrieving domain *y* variable for arbitrary range *x* variable:

    >>> x = 150
    >>> range_ = np.sin(np.linspace(0, 1, 10))
    >>> Signal(range_, domain)[x]  # doctest: +ELLIPSIS
    0.0359701...
    >>> x = np.linspace(100, 1000, 3)
    >>> Signal(range_, domain)[x]  # doctest: +ELLIPSIS
    array([  ...,   4.7669395...e-01,   8.4147098...e-01])

    Using an alternative interpolating function:

    >>> x = 150
    >>> from colour.algebra import CubicSplineInterpolator
    >>> Signal(
    ...     range_,
    ...     domain,
    ...     interpolator=CubicSplineInterpolator)[x]  # doctest: +ELLIPSIS
    0.0555274...
    >>> x = np.linspace(100, 1000, 3)
    >>> Signal(
    ...     range_,
    ...     domain,
    ...     interpolator=CubicSplineInterpolator)[x]  # doctest: +ELLIPSIS
    array([ 0.        ,  0.4794253...,  0.8414709...])
    """

    def __init__(
        self,
        data: Optional[Union[ArrayLike, dict, Series, Signal]] = None,
        domain: Optional[ArrayLike] = None,
        **kwargs: Any,
    ):
        super().__init__(kwargs.get("name"))

        self._dtype: Type[DTypeFloating] = DEFAULT_FLOAT_DTYPE
        self._domain: NDArray = np.array([])
        self._range: NDArray = np.array([])
        self._interpolator: Type[TypeInterpolator] = KernelInterpolator
        self._interpolator_kwargs: Dict = {}
        self._extrapolator: Type[TypeExtrapolator] = Extrapolator
        self._extrapolator_kwargs: Dict = {
            "method": "Constant",
            "left": np.nan,
            "right": np.nan,
        }

        self.domain, self.range = self.signal_unpack_data(data, domain)

        self.dtype = kwargs.get("dtype", self._dtype)

        self.interpolator = kwargs.get("interpolator", self._interpolator)
        self.interpolator_kwargs = kwargs.get(
            "interpolator_kwargs", self._interpolator_kwargs
        )
        self.extrapolator = kwargs.get("extrapolator", self._extrapolator)
        self.extrapolator_kwargs = kwargs.get(
            "extrapolator_kwargs", self._extrapolator_kwargs
        )

        self._create_function()

    @property
    def dtype(self) -> Type[DTypeFloating]:
        """
        Getter and setter property for the continuous signal dtype.

        Parameters
        ----------
        value
            Value to set the continuous signal dtype with.

        Returns
        -------
        DTypeFloating
            Continuous signal dtype.
        """

        return self._dtype

    @dtype.setter
    def dtype(self, value: Type[DTypeFloating]):
        """Setter for the **self.dtype** property."""

        attest(
            value in np.sctypes["float"],
            f'"dtype" must be one of the following types: '
            f"{np.sctypes['float']}",
        )

        self._dtype = value

        # The following self-assignments are written as intended and
        # triggers the rebuild of the underlying function.
        self.domain = self.domain
        self.range = self.range

    @property
    def domain(self) -> NDArray:
        """
        Getter and setter property for the continuous signal independent
        domain variable :math:`x`.

        Parameters
        ----------
        value
            Value to set the continuous signal independent domain
            variable :math:`x` with.

        Returns
        -------
        :class:`numpy.ndarray`
            Continuous signal independent domain variable :math:`x`.
        """

        return np.copy(self._domain)

    @domain.setter
    def domain(self, value: ArrayLike):
        """Setter for the **self.domain** property."""

        value = as_float_array(value, self.dtype)

        if not np.all(np.isfinite(value)):
            runtime_warning(
                f'"{self.name}" new "domain" variable is not finite: {value}, '
                f"unpredictable results may occur!"
            )

        if value.size != self._range.size:
            self._range = np.resize(self._range, value.shape)

        self._domain = value
        self._create_function()

    @property
    def range(self) -> NDArray:
        """
        Getter and setter property for the continuous signal corresponding
        range variable :math:`y`.

        Parameters
        ----------
        value
            Value to set the continuous signal corresponding range :math:`y`
            variable with.

        Returns
        -------
        :class:`numpy.ndarray`
            Continuous signal corresponding range variable :math:`y`.
        """

        return np.copy(self._range)

    @range.setter
    def range(self, value: ArrayLike):
        """Setter for the **self.range** property."""

        value = as_float_array(value, self.dtype)

        if not np.all(np.isfinite(value)):
            runtime_warning(
                f'"{self.name}" new "range" variable is not finite: {value}, '
                f"unpredictable results may occur!"
            )

        attest(
            value.size == self._domain.size,
            '"domain" and "range" variables must have same size!',
        )

        self._range = value
        self._create_function()

    @property
    def interpolator(self) -> Type[TypeInterpolator]:
        """
        Getter and setter property for the continuous signal interpolator type.

        Parameters
        ----------
        value
            Value to set the continuous signal interpolator type
            with.

        Returns
        -------
        Type[TypeInterpolator]
            Continuous signal interpolator type.
        """

        return self._interpolator

    @interpolator.setter
    def interpolator(self, value: Type[TypeInterpolator]):
        """Setter for the **self.interpolator** property."""

        # TODO: Check for interpolator compatibility.
        self._interpolator = value
        self._create_function()

    @property
    def interpolator_kwargs(self) -> Dict:
        """
        Getter and setter property for the continuous signal interpolator
        instantiation time arguments.

        Parameters
        ----------
        value
            Value to set the continuous signal interpolator instantiation
            time arguments to.

        Returns
        -------
        :class:`dict`
            Continuous signal interpolator instantiation time
            arguments.
        """

        return self._interpolator_kwargs

    @interpolator_kwargs.setter
    def interpolator_kwargs(self, value: dict):
        """Setter for the **self.interpolator_kwargs** property."""

        attest(
            isinstance(value, dict),
            f'"interpolator_kwargs" property: "{value}" type is not "dict"!',
        )

        self._interpolator_kwargs = value
        self._create_function()

    @property
    def extrapolator(self) -> Type[TypeExtrapolator]:
        """
        Getter and setter property for the continuous signal extrapolator type.

        Parameters
        ----------
        value
            Value to set the continuous signal extrapolator type
            with.

        Returns
        -------
        Type[TypeExtrapolator]
            Continuous signal extrapolator type.
        """

        return self._extrapolator

    @extrapolator.setter
    def extrapolator(self, value: Type[TypeExtrapolator]):
        """Setter for the **self.extrapolator** property."""

        # TODO: Check for extrapolator compatibility.
        self._extrapolator = value
        self._create_function()

    @property
    def extrapolator_kwargs(self) -> Dict:
        """
        Getter and setter property for the continuous signal extrapolator
        instantiation time arguments.

        Parameters
        ----------
        value
            Value to set the continuous signal extrapolator instantiation
            time arguments to.

        Returns
        -------
        :class:`dict`
            Continuous signal extrapolator instantiation time
            arguments.
        """

        return self._extrapolator_kwargs

    @extrapolator_kwargs.setter
    def extrapolator_kwargs(self, value: dict):
        """Setter for the **self.extrapolator_kwargs** property."""

        attest(
            isinstance(value, dict),
            f'"extrapolator_kwargs" property: "{value}" type is not "dict"!',
        )

        self._extrapolator_kwargs = value
        self._create_function()

    @property
    def function(self) -> Callable:
        """
        Getter property for the continuous signal callable.

        Returns
        -------
        Callable
            Continuous signal callable.
        """

        return self._function

    def __str__(self) -> str:
        """
        Return a formatted string representation of the continuous signal.

        Returns
        -------
        :class:`str`
            Formatted string representation.

        Examples
        --------
        >>> range_ = np.linspace(10, 100, 10)
        >>> print(Signal(range_))
        [[   0.   10.]
         [   1.   20.]
         [   2.   30.]
         [   3.   40.]
         [   4.   50.]
         [   5.   60.]
         [   6.   70.]
         [   7.   80.]
         [   8.   90.]
         [   9.  100.]]
        """

        try:
            return str(tstack([self.domain, self.range]))
        except TypeError:
            return super().__str__()

    def __repr__(self) -> str:
        """
        Return an evaluable string representation of the continuous signal.

        Returns
        -------
        :class:`str`
            Evaluable string representation.

        Examples
        --------
        >>> range_ = np.linspace(10, 100, 10)
        >>> Signal(range_)  # doctest: +ELLIPSIS
        Signal([[   0.,   10.],
                [   1.,   20.],
                [   2.,   30.],
                [   3.,   40.],
                [   4.,   50.],
                [   5.,   60.],
                [   6.,   70.],
                [   7.,   80.],
                [   8.,   90.],
                [   9.,  100.]],
               interpolator=KernelInterpolator,
               interpolator_kwargs={},
               extrapolator=Extrapolator,
               extrapolator_kwargs={...})
        """

        if is_documentation_building():  # pragma: no cover
            return f"{self.__class__.__name__}(name='{self.name}', ...)"

        try:
            representation = repr(tstack([self.domain, self.range]))
            representation = representation.replace(
                "array", self.__class__.__name__
            )
            representation = representation.replace(
                "       [",
                f"{' ' * (len(self.__class__.__name__) + 2)}[",
            )
            indentation = " " * (len(self.__class__.__name__) + 1)
            representation = (
                f"{representation[:-1]},\n"
                f"{indentation}interpolator={self.interpolator.__name__},\n"
                f"{indentation}interpolator_kwargs="
                f"{repr(self.interpolator_kwargs)},\n"
                f"{indentation}extrapolator={self.extrapolator.__name__},\n"
                f"{indentation}extrapolator_kwargs="
                f"{repr(self.extrapolator_kwargs)})"
            )

            return representation
        except TypeError:
            return super().__repr__()

    def __hash__(self) -> Integer:
        """
        Return the abstract continuous function hash.

        Returns
        -------
        :class:`numpy.integer`
            Object hash.
        """

        return hash(
            (
                self.domain.tobytes(),
                self.range.tobytes(),
                self.interpolator.__name__,
                repr(self.interpolator_kwargs),
                self.extrapolator.__name__,
                repr(self.extrapolator_kwargs),
            )
        )

    def __getitem__(
        self, x: Union[FloatingOrArrayLike, slice]
    ) -> FloatingOrNDArray:
        """
        Return the corresponding range variable :math:`y` for independent
        domain variable :math:`x`.

        Parameters
        ----------
        x
            Independent domain variable :math:`x`.

        Returns
        -------
        :class:`numpy.floating` or :class:`numpy.ndarray`
            Variable :math:`y` range value.

        Examples
        --------
        >>> range_ = np.linspace(10, 100, 10)
        >>> signal = Signal(range_)
        >>> print(signal)
        [[   0.   10.]
         [   1.   20.]
         [   2.   30.]
         [   3.   40.]
         [   4.   50.]
         [   5.   60.]
         [   6.   70.]
         [   7.   80.]
         [   8.   90.]
         [   9.  100.]]
        >>> signal[0]
        10.0
        >>> signal[np.array([0, 1, 2])]
        array([ 10.,  20.,  30.])
        >>> signal[0:3]
        array([ 10.,  20.,  30.])
        >>> signal[np.linspace(0, 5, 5)]  # doctest: +ELLIPSIS
        array([ 10.        ,  22.8348902...,  34.8004492...,  \
47.5535392...,  60.        ])
        """

        if isinstance(x, slice):
            return self._range[x]
        else:
            return self._function(x)

    def __setitem__(
        self, x: Union[FloatingOrArrayLike, slice], y: FloatingOrArrayLike
    ):
        """
        Set the corresponding range variable :math:`y` for independent domain
        variable :math:`x`.

        Parameters
        ----------
        x
            Independent domain variable :math:`x`.
        y
            Corresponding range variable :math:`y`.

        Examples
        --------
        >>> range_ = np.linspace(10, 100, 10)
        >>> signal = Signal(range_)
        >>> print(signal)
        [[   0.   10.]
         [   1.   20.]
         [   2.   30.]
         [   3.   40.]
         [   4.   50.]
         [   5.   60.]
         [   6.   70.]
         [   7.   80.]
         [   8.   90.]
         [   9.  100.]]
        >>> signal[0] = 20
        >>> signal[0]
        20.0
        >>> signal[np.array([0, 1, 2])] = 30
        >>> signal[np.array([0, 1, 2])]
        array([ 30.,  30.,  30.])
        >>> signal[0:3] = 40
        >>> signal[0:3]
        array([ 40.,  40.,  40.])
        >>> signal[np.linspace(0, 5, 5)] = 50
        >>> print(signal)
        [[   0.     50.  ]
         [   1.     40.  ]
         [   1.25   50.  ]
         [   2.     40.  ]
         [   2.5    50.  ]
         [   3.     40.  ]
         [   3.75   50.  ]
         [   4.     50.  ]
         [   5.     50.  ]
         [   6.     70.  ]
         [   7.     80.  ]
         [   8.     90.  ]
         [   9.    100.  ]]
        >>> signal[np.array([0, 1, 2])] = np.array([10, 20, 30])
        >>> print(signal)
        [[   0.     10.  ]
         [   1.     20.  ]
         [   1.25   50.  ]
         [   2.     30.  ]
         [   2.5    50.  ]
         [   3.     40.  ]
         [   3.75   50.  ]
         [   4.     50.  ]
         [   5.     50.  ]
         [   6.     70.  ]
         [   7.     80.  ]
         [   8.     90.  ]
         [   9.    100.  ]]
        """

        if isinstance(x, slice):
            self._range[x] = y
        else:
            x = np.atleast_1d(x).astype(self.dtype)
            y = np.resize(y, x.shape)

            # Matching domain, updating existing `self._range` values.
            mask = np.in1d(x, self._domain)
            x_m = x[mask]
            indexes = np.searchsorted(self._domain, x_m)
            self._range[indexes] = y[mask]

            # Non matching domain, inserting into existing `self.domain`
            # and `self.range`.
            x_nm = x[~mask]
            indexes = np.searchsorted(self._domain, x_nm)
            if indexes.size != 0:
                self._domain = np.insert(self._domain, indexes, x_nm)
                self._range = np.insert(self._range, indexes, y[~mask])

        self._create_function()

    def __contains__(self, x: Union[FloatingOrArrayLike, slice]) -> bool:
        """
        Return whether the continuous signal contains given independent domain
        variable :math:`x`.

        Parameters
        ----------
        x
            Independent domain variable :math:`x`.

        Returns
        -------
        :class:`bool`
            Whether :math:`x` domain value is contained.

        Examples
        --------
        >>> range_ = np.linspace(10, 100, 10)
        >>> signal = Signal(range_)
        >>> 0 in signal
        True
        >>> 0.5 in signal
        True
        >>> 1000 in signal
        False
        """

        return bool(
            np.all(
                np.where(
                    np.logical_and(
                        x >= np.min(self._domain), x <= np.max(self._domain)
                    ),
                    True,
                    False,
                )
            )
        )

    def __eq__(self, other: Any) -> bool:
        """
        Return whether the continuous signal is equal to given other object.

        Parameters
        ----------
        other
            Object to test whether it is equal to the continuous signal.

        Returns
        -------
        :class:`bool`
            Whether given object is equal to the continuous signal.

        Examples
        --------
        >>> range_ = np.linspace(10, 100, 10)
        >>> signal_1 = Signal(range_)
        >>> signal_2 = Signal(range_)
        >>> signal_1 == signal_2
        True
        >>> signal_2[0] = 20
        >>> signal_1 == signal_2
        False
        >>> signal_2[0] = 10
        >>> signal_1 == signal_2
        True
        >>> from colour.algebra import CubicSplineInterpolator
        >>> signal_2.interpolator = CubicSplineInterpolator
        >>> signal_1 == signal_2
        False
        """

        if isinstance(other, Signal):
            return all(
                [
                    np.array_equal(self._domain, other.domain),
                    np.array_equal(self._range, other.range),
                    self._interpolator is other.interpolator,
                    self._interpolator_kwargs == other.interpolator_kwargs,
                    self._extrapolator is other.extrapolator,
                    self._extrapolator_kwargs == other.extrapolator_kwargs,
                ]
            )
        else:
            return False

    def __ne__(self, other: Any) -> bool:
        """
        Return whether the continuous signal is not equal to given other
        object.

        Parameters
        ----------
        other
            Object to test whether it is not equal to the continuous signal.

        Returns
        -------
        :class:`bool`
            Whether given object is not equal to the continuous signal.

        Examples
        --------
        >>> range_ = np.linspace(10, 100, 10)
        >>> signal_1 = Signal(range_)
        >>> signal_2 = Signal(range_)
        >>> signal_1 != signal_2
        False
        >>> signal_2[0] = 20
        >>> signal_1 != signal_2
        True
        >>> signal_2[0] = 10
        >>> signal_1 != signal_2
        False
        >>> from colour.algebra import CubicSplineInterpolator
        >>> signal_2.interpolator = CubicSplineInterpolator
        >>> signal_1 != signal_2
        True
        """

        return not (self == other)

    def _create_function(self):
        """Create the continuous signal underlying function."""

        if self._domain.size != 0 and self._range.size != 0:
            self._function = self._extrapolator(
                self._interpolator(
                    self.domain, self.range, **self._interpolator_kwargs
                ),
                **self._extrapolator_kwargs,
            )
        else:

            def _undefined_function(*args: Any, **kwargs: Any):
                """
                Raise a :class:`ValueError` exception.

                Other Parameters
                ----------------
                args
                    Arguments.
                kwargs
                    Keywords arguments.

                Raises
                ------
                ValueError
                """

                raise ValueError(
                    "Underlying signal interpolator function does not exists, "
                    "please ensure you defined both "
                    '"domain" and "range" variables!'
                )

            self._function = _undefined_function

    def _fill_domain_nan(
        self,
        method: Union[
            Literal["Constant", "Interpolation"], str
        ] = "Interpolation",
        default: Number = 0,
    ):
        """
        Fill NaNs in independent domain variable :math:`x` using given method.

        Parameters
        ----------
        method
            *Interpolation* method linearly interpolates through the NaNs,
            *Constant* method replaces NaNs with ``default``.
        default
            Value to use with the *Constant* method.

        Returns
        -------
        :class:`colour.continuous.Signal`
            NaNs filled continuous signal independent domain :math:`x`
            variable.
        """

        self._domain = fill_nan(self._domain, method, default)
        self._create_function()

    def _fill_range_nan(
        self,
        method: Union[
            Literal["Constant", "Interpolation"], str
        ] = "Interpolation",
        default: Number = 0,
    ):
        """
        Fill NaNs in corresponding range variable :math:`y` using given method.

        Parameters
        ----------
        method
            *Interpolation* method linearly interpolates through the NaNs,
            *Constant* method replaces NaNs with ``default``.
        default
            Value to use with the *Constant* method.

        Returns
        -------
        :class:`colour.continuous.Signal`
            NaNs filled continuous signal i corresponding range :math:`y`
            variable.
        """

        self._range = fill_nan(self._range, method, default)
        self._create_function()

    def arithmetical_operation(
        self,
        a: Union[FloatingOrArrayLike, AbstractContinuousFunction],
        operation: Literal["+", "-", "*", "/", "**"],
        in_place: Boolean = False,
    ) -> AbstractContinuousFunction:
        """
        Perform given arithmetical operation with operand :math:`a`, the
        operation can be either performed on a copy or in-place.

        Parameters
        ----------
        a
            Operand :math:`a`.
        operation
            Operation to perform.
        in_place
            Operation happens in place.

        Returns
        -------
        :class:`colour.continuous.Signal`
            Continuous signal.

        Examples
        --------
        Adding a single *numeric* variable:

        >>> range_ = np.linspace(10, 100, 10)
        >>> signal_1 = Signal(range_)
        >>> print(signal_1)
        [[   0.   10.]
         [   1.   20.]
         [   2.   30.]
         [   3.   40.]
         [   4.   50.]
         [   5.   60.]
         [   6.   70.]
         [   7.   80.]
         [   8.   90.]
         [   9.  100.]]
        >>> print(signal_1.arithmetical_operation(10, '+', True))
        [[   0.   20.]
         [   1.   30.]
         [   2.   40.]
         [   3.   50.]
         [   4.   60.]
         [   5.   70.]
         [   6.   80.]
         [   7.   90.]
         [   8.  100.]
         [   9.  110.]]

        Adding an `ArrayLike` variable:

        >>> a = np.linspace(10, 100, 10)
        >>> print(signal_1.arithmetical_operation(a, '+', True))
        [[   0.   30.]
         [   1.   50.]
         [   2.   70.]
         [   3.   90.]
         [   4.  110.]
         [   5.  130.]
         [   6.  150.]
         [   7.  170.]
         [   8.  190.]
         [   9.  210.]]

        Adding a :class:`colour.continuous.Signal` class:

        >>> signal_2 = Signal(range_)
        >>> print(signal_1.arithmetical_operation(signal_2, '+', True))
        [[   0.   40.]
         [   1.   70.]
         [   2.  100.]
         [   3.  130.]
         [   4.  160.]
         [   5.  190.]
         [   6.  220.]
         [   7.  250.]
         [   8.  280.]
         [   9.  310.]]
        """

        operator, ioperator = {
            "+": (add, iadd),
            "-": (sub, isub),
            "*": (mul, imul),
            "/": (truediv, itruediv),
            "**": (pow, ipow),
        }[operation]

        if in_place:
            if isinstance(a, Signal):
                self[self._domain] = operator(self._range, a[self._domain])
                exclusive_or = np.setxor1d(self._domain, a.domain)
                self[exclusive_or] = full(exclusive_or.shape, np.nan)
            else:
                self.range = ioperator(self.range, a)

            return self
        else:
            copy = ioperator(self.copy(), a)

            return copy

    @staticmethod
    def signal_unpack_data(
        data=Optional[Union[ArrayLike, dict, Series, "Signal"]],
        domain: Optional[ArrayLike] = None,
        dtype: Optional[Type[DTypeFloating]] = None,
    ) -> Tuple:
        """
        Unpack given data for continuous signal instantiation.

        Parameters
        ----------
        data
            Data to unpack for continuous signal instantiation.
        domain
            Values to initialise the :attr:`colour.continuous.Signal.domain`
            attribute with. If both ``data`` and ``domain`` arguments are
            defined, the latter will be used to initialise the
            :attr:`colour.continuous.Signal.domain` property.
        dtype
            Floating point data type.

        Returns
        -------
        :class:`tuple`
            Independent domain variable :math:`x` and corresponding range
            variable :math:`y` unpacked for continuous signal instantiation.

        Examples
        --------
        Unpacking using implicit *domain*:

        >>> range_ = np.linspace(10, 100, 10)
        >>> domain, range_ = Signal.signal_unpack_data(range_)
        >>> print(domain)
        [ 0.  1.  2.  3.  4.  5.  6.  7.  8.  9.]
        >>> print(range_)
        [  10.   20.   30.   40.   50.   60.   70.   80.   90.  100.]

        Unpacking using explicit *domain*:

        >>> domain = np.arange(100, 1100, 100)
        >>> domain, range = Signal.signal_unpack_data(range_, domain)
        >>> print(domain)
        [  100.   200.   300.   400.   500.   600.   700.   800.   900.  1000.]
        >>> print(range_)
        [  10.   20.   30.   40.   50.   60.   70.   80.   90.  100.]

        Unpacking using a *dict*:

        >>> domain, range_ = Signal.signal_unpack_data(
        ...     dict(zip(domain, range_)))
        >>> print(domain)
        [  100.   200.   300.   400.   500.   600.   700.   800.   900.  1000.]
        >>> print(range_)
        [  10.   20.   30.   40.   50.   60.   70.   80.   90.  100.]

        Unpacking using a *Pandas* :class:`pandas.Series`:

        >>> if is_pandas_installed():
        ...     from pandas import Series
        ...     domain, range = Signal.signal_unpack_data(
        ...         Series(dict(zip(domain, range_))))
        ... # doctest: +ELLIPSIS
        >>> print(domain)  # doctest: +SKIP
        [  100.   200.   300.   400.   500.   600.   700.   800.   900.  1000.]
        >>> print(range_)  # doctest: +SKIP
        [  10.   20.   30.   40.   50.   60.   70.   80.   90.  100.]

        Unpacking using a :class:`colour.continuous.Signal` class:

        >>> domain, range_ = Signal.signal_unpack_data(
        ...     Signal(range_, domain))
        >>> print(domain)
        [  100.   200.   300.   400.   500.   600.   700.   800.   900.  1000.]
        >>> print(range_)
        [  10.   20.   30.   40.   50.   60.   70.   80.   90.  100.]
        """

        dtype = cast(Type[DTypeFloating], optional(dtype, DEFAULT_FLOAT_DTYPE))

        domain_unpacked: NDArray = np.array([])
        range_unpacked: NDArray = np.array([])

        if isinstance(data, Signal):
            domain_unpacked = data.domain
            range_unpacked = data.range
        elif issubclass(type(data), Sequence) or isinstance(
            data, (tuple, list, np.ndarray, Iterator, ValuesView)
        ):
            data_array = tsplit(list(data))

            attest(data_array.ndim == 1, 'User "data" must be 1-dimensional!')

            domain_unpacked, range_unpacked = (
                np.arange(0, data_array.size, dtype=dtype),
                data_array,
            )
        elif issubclass(type(data), Mapping) or isinstance(data, dict):
            domain_unpacked, range_unpacked = tsplit(sorted(data.items()))
        elif is_pandas_installed():
            if isinstance(data, Series):
                domain_unpacked = data.index.values
                range_unpacked = data.values

        if domain is not None:
            domain_array = as_float_array(list(domain), dtype)  # type: ignore[arg-type]

            attest(
                len(domain_array) == len(range_unpacked),
                'User "domain" length is not compatible with unpacked "range"!',
            )

            domain_unpacked = domain_array

        if range_unpacked is not None:
            range_unpacked = as_float_array(range_unpacked, dtype)

        return domain_unpacked, range_unpacked

    def fill_nan(
        self,
        method: Union[
            Literal["Constant", "Interpolation"], str
        ] = "Interpolation",
        default: Number = 0,
    ) -> AbstractContinuousFunction:
        """
        Fill NaNs in independent domain variable :math:`x` and corresponding
        range variable :math:`y` using given method.

        Parameters
        ----------
        method
            *Interpolation* method linearly interpolates through the NaNs,
            *Constant* method replaces NaNs with ``default``.
        default
            Value to use with the *Constant* method.

        Returns
        -------
        :class:`colour.continuous.Signal`
            NaNs filled continuous signal.

        Examples
        --------
        >>> range_ = np.linspace(10, 100, 10)
        >>> signal = Signal(range_)
        >>> signal[3:7] = np.nan
        >>> print(signal)
        [[   0.   10.]
         [   1.   20.]
         [   2.   30.]
         [   3.   nan]
         [   4.   nan]
         [   5.   nan]
         [   6.   nan]
         [   7.   80.]
         [   8.   90.]
         [   9.  100.]]
        >>> print(signal.fill_nan())
        [[   0.   10.]
         [   1.   20.]
         [   2.   30.]
         [   3.   40.]
         [   4.   50.]
         [   5.   60.]
         [   6.   70.]
         [   7.   80.]
         [   8.   90.]
         [   9.  100.]]
        >>> signal[3:7] = np.nan
        >>> print(signal.fill_nan(method='Constant'))
        [[   0.   10.]
         [   1.   20.]
         [   2.   30.]
         [   3.    0.]
         [   4.    0.]
         [   5.    0.]
         [   6.    0.]
         [   7.   80.]
         [   8.   90.]
         [   9.  100.]]
        """

        method = validate_method(method, ["Interpolation", "Constant"])

        self._fill_domain_nan(method, default)
        self._fill_range_nan(method, default)

        return self

    @required("Pandas")
    def to_series(self) -> Series:
        """
        Convert the continuous signal to a *Pandas* :class:`pandas.Series`
        class instance.

        Returns
        -------
        :class:`pandas.Series`
            Continuous signal as a *Pandas*:class:`pandas.Series` class
            instance.

        Examples
        --------
        >>> if is_pandas_installed():
        ...     range_ = np.linspace(10, 100, 10)
        ...     signal = Signal(range_)
        ...     print(signal.to_series())  # doctest: +SKIP
        0.0     10.0
        1.0     20.0
        2.0     30.0
        3.0     40.0
        4.0     50.0
        5.0     60.0
        6.0     70.0
        7.0     80.0
        8.0     90.0
        9.0    100.0
        Name: Signal (...), dtype: float64
        """

        return Series(data=self._range, index=self._domain, name=self.name)



"""Showcases colour models plotting examples."""

import numpy as np
from pprint import pprint

import colour
from colour.plotting import (
    plot_RGB_colourspaces_in_chromaticity_diagram_CIE1931,
    plot_RGB_colourspaces_in_chromaticity_diagram_CIE1960UCS,
    plot_RGB_colourspaces_in_chromaticity_diagram_CIE1976UCS,
    plot_RGB_chromaticities_in_chromaticity_diagram_CIE1931,
    plot_RGB_chromaticities_in_chromaticity_diagram_CIE1960UCS,
    plot_RGB_chromaticities_in_chromaticity_diagram_CIE1976UCS,
    colour_style,
    plot_multi_cctfs,
    plot_single_cctf,
)
from colour.utilities import message_box

message_box("Colour Models Plots")

colour_style()

message_box(
    'Plotting "RGB" colourspaces in the ' '"CIE 1931 Chromaticity Diagram".'
)
pprint(sorted(colour.RGB_COLOURSPACES.keys()))
plot_RGB_colourspaces_in_chromaticity_diagram_CIE1931(
    ["ITU-R BT.709", "ACEScg", "S-Gamut"], show_pointer_gamut=True
)

print("\n")

message_box(
    'Plotting "RGB" colourspaces in the '
    '"CIE 1960 UCS Chromaticity Diagram".'
)
pprint(sorted(colour.RGB_COLOURSPACES.keys()))
plot_RGB_colourspaces_in_chromaticity_diagram_CIE1960UCS(
    ["ITU-R BT.709", "ACEScg", "S-Gamut"], show_pointer_gamut=True
)

print("\n")

message_box(
    'Plotting "RGB" colourspaces in the '
    '"CIE 1976 UCS Chromaticity Diagram".'
)
pprint(sorted(colour.RGB_COLOURSPACES.keys()))
plot_RGB_colourspaces_in_chromaticity_diagram_CIE1976UCS(
    ["ITU-R BT.709", "ACEScg", "S-Gamut"], show_pointer_gamut=True
)

print("\n")

RGB = np.random.random((32, 32, 3))

message_box(
    'Plotting "RGB" chromaticity coordinates in the '
    '"CIE 1931 Chromaticity Diagram".'
)
plot_RGB_chromaticities_in_chromaticity_diagram_CIE1931(
    RGB,
    "ITU-R BT.709",
    colourspaces=["ACEScg", "S-Gamut"],
    show_pointer_gamut=True,
)

print("\n")

message_box(
    'Plotting "RGB" chromaticity coordinates in the '
    '"CIE 1960 UCS Chromaticity Diagram".'
)
plot_RGB_chromaticities_in_chromaticity_diagram_CIE1960UCS(
    RGB,
    "ITU-R BT.709",
    colourspaces=["ACEScg", "S-Gamut"],
    show_pointer_gamut=True,
)

print("\n")

message_box(
    'Plotting "RGB" chromaticity coordinates in the '
    '"CIE 1976 UCS Chromaticity Diagram".'
)
plot_RGB_chromaticities_in_chromaticity_diagram_CIE1976UCS(
    RGB,
    "ITU-R BT.709",
    colourspaces=["ACEScg", "S-Gamut"],
    show_pointer_gamut=True,
)

print("\n")

message_box(
    'Plotting a single custom "RGB" colourspace in the '
    '"CIE 1931 Chromaticity Diagram".'
)
AWFUL_RGB = colour.RGB_Colourspace(
    "Awful RGB",
    primaries=np.array(
        [
            [0.10, 0.20],
            [0.30, 0.15],
            [0.05, 0.60],
        ]
    ),
    whitepoint=np.array([1.0 / 3.0, 1.0 / 3.0]),
)
pprint(sorted(colour.RGB_COLOURSPACES.keys()))
plot_RGB_colourspaces_in_chromaticity_diagram_CIE1931(
    ["ITU-R BT.709", AWFUL_RGB]
)

print("\n")

message_box(
    'Plotting a single "RGB" colourspace encoding colour component transfer '
    "function."
)
plot_single_cctf("ITU-R BT.709")

print("\n")

message_box(
    'Plotting multiple "RGB" colourspaces encoding colour component transfer '
    "functions."
)
plot_multi_cctfs(["ITU-R BT.709", "sRGB"])

message_box(
    'Plotting multiple "RGB" colourspaces decoding colour component transfer '
    "functions."
)
plot_multi_cctfs(["ACES2065-1", "ProPhoto RGB"], cctf_decoding=True)



from .breneman1987 import (
    BRENEMAN_EXPERIMENTS,
    BRENEMAN_EXPERIMENT_PRIMARIES_CHROMATICITIES,
)

__all__ = [
    "BRENEMAN_EXPERIMENTS",
    "BRENEMAN_EXPERIMENT_PRIMARIES_CHROMATICITIES",
]



"""
Verbose
=======

Defines the verbose related objects.
"""

from __future__ import annotations

import numpy as np
import os
import sys
import traceback
import warnings
from collections import defaultdict
from contextlib import contextmanager
from itertools import chain
from textwrap import TextWrapper
from warnings import filterwarnings, formatwarning, warn

from colour.utilities import is_string, optional
from colour.hints import (
    Any,
    Boolean,
    Callable,
    Dict,
    Integer,
    LiteralWarning,
    Mapping,
    Generator,
    Optional,
    TextIO,
    Type,
    Union,
    cast,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "ColourWarning",
    "ColourUsageWarning",
    "ColourRuntimeWarning",
    "message_box",
    "show_warning",
    "warning",
    "runtime_warning",
    "usage_warning",
    "filter_warnings",
    "suppress_warnings",
    "numpy_print_options",
    "ANCILLARY_COLOUR_SCIENCE_PACKAGES",
    "ANCILLARY_RUNTIME_PACKAGES",
    "ANCILLARY_DEVELOPMENT_PACKAGES",
    "ANCILLARY_EXTRAS_PACKAGES",
    "describe_environment",
]


class ColourWarning(Warning):
    """
    Define the base class of *Colour* warnings.

    It is a subclass of the :class:`Warning` class.
    """


class ColourUsageWarning(Warning):
    """
    Define the base class of *Colour* usage warnings.

    It is a subclass of the :class:`colour.utilities.ColourWarning` class.
    """


class ColourRuntimeWarning(Warning):
    """
    Define the base class of *Colour* runtime warnings.

    It is a subclass of the :class:`colour.utilities.ColourWarning` class.
    """


def message_box(
    message: str,
    width: Integer = 79,
    padding: Integer = 3,
    print_callable: Callable = print,
):
    """
    Print a message inside a box.

    Parameters
    ----------
    message
        Message to print.
    width
        Message box width.
    padding
        Padding on each side of the message.
    print_callable
        Callable used to print the message box.

    Examples
    --------
    >>> message = (
    ...     'Lorem ipsum dolor sit amet, consectetur adipiscing elit, '
    ...     'sed do eiusmod tempor incididunt ut labore et dolore magna '
    ...     'aliqua.')
    >>> message_box(message, width=75)
    ===========================================================================
    *                                                                         *
    *   Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do       *
    *   eiusmod tempor incididunt ut labore et dolore magna aliqua.           *
    *                                                                         *
    ===========================================================================
    >>> message_box(message, width=60)
    ============================================================
    *                                                          *
    *   Lorem ipsum dolor sit amet, consectetur adipiscing     *
    *   elit, sed do eiusmod tempor incididunt ut labore et    *
    *   dolore magna aliqua.                                   *
    *                                                          *
    ============================================================
    >>> message_box(message, width=75, padding=16)
    ===========================================================================
    *                                                                         *
    *                Lorem ipsum dolor sit amet, consectetur                  *
    *                adipiscing elit, sed do eiusmod tempor                   *
    *                incididunt ut labore et dolore magna                     *
    *                aliqua.                                                  *
    *                                                                         *
    ===========================================================================
    """

    ideal_width = width - padding * 2 - 2

    def inner(text):
        """Format and pads inner text for the message box."""

        return (
            f'*{" " * padding}'
            f'{text}{" " * (width - len(text) - padding * 2 - 2)}'
            f'{" " * padding}*'
        )

    print_callable("=" * width)
    print_callable(inner(""))

    wrapper = TextWrapper(
        width=ideal_width, break_long_words=False, replace_whitespace=False
    )

    lines = [wrapper.wrap(line) for line in message.split("\n")]
    for line in chain(*[" " if len(line) == 0 else line for line in lines]):
        print_callable(inner(line.expandtabs()))

    print_callable(inner(""))
    print_callable("=" * width)


def show_warning(
    message: Union[Warning, str],
    category: Type[Warning],
    filename: str,
    lineno: Integer,
    file: Optional[TextIO] = None,
    line: Optional[str] = None,
) -> None:
    """
    Alternative :func:`warnings.showwarning` definition that allows traceback
    printing.

    This definition is expected to be used by setting the
    *COLOUR_SCIENCE__COLOUR__SHOW_WARNINGS_WITH_TRACEBACK* environment variable
    prior to importing *colour*.

    Parameters
    ----------
    message
        Warning message.
    category
        :class:`Warning` sub-class.
    filename
        File path to read the line at ``lineno`` from if ``line`` is None.
    lineno
        Line number to read the line at in ``filename`` if ``line`` is None.
    file
        :class:`file` object to write the warning to, defaults to
        :attr:`sys.stderr` attribute.
    line
        Source code to be included in the warning message.

    Notes
    -----
    -   Setting the *COLOUR_SCIENCE__COLOUR__SHOW_WARNINGS_WITH_TRACEBACK*
        environment variable will result in the :func:`warnings.showwarning`
        definition to be replaced with the
        :func:`colour.utilities.show_warning` definition and thus providing
        complete traceback from the point where the warning occurred.
    """

    frame_range = (1, None)

    file = optional(file, sys.stderr)
    if file is None:
        return

    try:
        # Generating a traceback to print useful warning origin.
        frame_in, frame_out = frame_range

        try:
            raise ZeroDivisionError
        except ZeroDivisionError:
            exception_traceback = sys.exc_info()[2]
            frame = (
                exception_traceback.tb_frame.f_back
                if exception_traceback is not None
                else None
            )
            while frame_in and frame is not None:
                frame = frame.f_back
                frame_in -= 1

        traceback.print_stack(frame, frame_out, file)

        file.write(formatwarning(message, category, filename, lineno, line))
    except (OSError, UnicodeError):
        pass


if os.environ.get(  # pragma: no cover
    "COLOUR_SCIENCE__COLOUR__SHOW_WARNINGS_WITH_TRACEBACK"
):
    warnings.showwarning = show_warning  # pragma: no cover


def warning(*args: Any, **kwargs: Any):
    """
    Issue a warning.

    Other Parameters
    ----------------
    args
        Arguments.
    kwargs
        Keywords arguments.

    Examples
    --------
    >>> warning('This is a warning!')  # doctest: +SKIP
    """

    kwargs["category"] = kwargs.get("category", ColourWarning)

    warn(*args, **kwargs)


def runtime_warning(*args: Any, **kwargs: Any):
    """
    Issue a runtime warning.

    Other Parameters
    ----------------
    args
        Arguments.
    kwargs
        Keywords arguments.

    Examples
    --------
    >>> usage_warning('This is a runtime warning!')  # doctest: +SKIP
    """

    kwargs["category"] = ColourRuntimeWarning

    warning(*args, **kwargs)


def usage_warning(*args: Any, **kwargs: Any):
    """
    Issue a usage warning.

    Other Parameters
    ----------------
    args
        Arguments.
    kwargs
        Keywords arguments.

    Examples
    --------
    >>> usage_warning('This is an usage warning!')  # doctest: +SKIP
    """

    kwargs["category"] = ColourUsageWarning

    warning(*args, **kwargs)


def filter_warnings(
    colour_runtime_warnings: Optional[Union[bool, LiteralWarning]] = None,
    colour_usage_warnings: Optional[Union[bool, LiteralWarning]] = None,
    colour_warnings: Optional[Union[bool, LiteralWarning]] = None,
    python_warnings: Optional[Union[bool, LiteralWarning]] = None,
):
    """
    Filter *Colour* and also optionally overall Python warnings.

    The possible values for all the actions, i.e. each argument, are as
    follows:

    - *None* (No action is taken)
    - *True* (*ignore*)
    - *False* (*default*)
    - *error*
    - *ignore*
    - *always*
    - *default*
    - *module*
    - *once*

    Parameters
    ----------
    colour_runtime_warnings
        Whether to filter *Colour* runtime warnings according to the action
        value.
    colour_usage_warnings
        Whether to filter *Colour* usage warnings according to the action
        value.
    colour_warnings
        Whether to filter *Colour* warnings, this also filters *Colour* usage
        and runtime warnings according to the action value.
    python_warnings
        Whether to filter *Python* warnings  according to the action value.

    Examples
    --------
    Filtering *Colour* runtime warnings:

    >>> filter_warnings(colour_runtime_warnings=True)

    Filtering *Colour* usage warnings:

    >>> filter_warnings(colour_usage_warnings=True)

    Filtering *Colour* warnings:

    >>> filter_warnings(colour_warnings=True)

    Filtering all the *Colour* and also Python warnings:

    >>> filter_warnings(python_warnings=True)

    Enabling all the *Colour* and Python warnings:

    >>> filter_warnings(*[False] * 4)

    Enabling all the *Colour* and Python warnings using the *default* action:

    >>> filter_warnings(*['default'] * 4)

    Setting back the default state:

    >>> filter_warnings(colour_runtime_warnings=True)
    """

    for action, category in [
        (colour_warnings, ColourWarning),
        (colour_runtime_warnings, ColourRuntimeWarning),
        (colour_usage_warnings, ColourUsageWarning),
        (python_warnings, Warning),
    ]:
        if action is None:
            continue

        if is_string(action):
            action = cast(LiteralWarning, str(action))
        else:
            action = "ignore" if action else "default"

        filterwarnings(action, category=category)


# Defaulting to filter *Colour* runtime warnings.
filter_warnings(colour_runtime_warnings=True)


@contextmanager
def suppress_warnings(
    colour_runtime_warnings: Optional[Union[bool, LiteralWarning]] = None,
    colour_usage_warnings: Optional[Union[bool, LiteralWarning]] = None,
    colour_warnings: Optional[Union[bool, LiteralWarning]] = None,
    python_warnings: Optional[Union[bool, LiteralWarning]] = None,
) -> Generator:
    """
    Define a context manager filtering *Colour* and also optionally overall
    Python warnings.

    The possible values for all the actions, i.e. each argument, are as
    follows:

    - *None* (No action is taken)
    - *True* (*ignore*)
    - *False* (*default*)
    - *error*
    - *ignore*
    - *always*
    - *default*
    - *module*
    - *once*

    Parameters
    ----------
    colour_runtime_warnings
        Whether to filter *Colour* runtime warnings according to the action
        value.
    colour_usage_warnings
        Whether to filter *Colour* usage warnings according to the action
        value.
    colour_warnings
        Whether to filter *Colour* warnings, this also filters *Colour* usage
        and runtime warnings according to the action value.
    python_warnings
        Whether to filter *Python* warnings  according to the action value.
    """

    filters = warnings.filters
    show_warnings = warnings.showwarning

    filter_warnings(
        colour_warnings=colour_warnings,
        colour_runtime_warnings=colour_runtime_warnings,
        colour_usage_warnings=colour_usage_warnings,
        python_warnings=python_warnings,
    )

    try:
        yield
    finally:
        warnings.filters = filters
        warnings.showwarning = show_warnings


@contextmanager
def numpy_print_options(*args: Any, **kwargs: Any) -> Generator:
    """
    Define a context manager implementing context changes to *Numpy* print
    behaviour.

    Other Parameters
    ----------------
    args
        Arguments.
    kwargs
        Keywords arguments.

    Examples
    --------
    >>> np.array([np.pi])  # doctest: +ELLIPSIS
    array([ 3.1415926...])
    >>> with numpy_print_options(formatter={'float': '{:0.1f}'.format}):
    ...      np.array([np.pi])
    array([3.1])
    """

    options = np.get_printoptions()
    np.set_printoptions(*args, **kwargs)
    try:
        yield
    finally:
        np.set_printoptions(**options)


ANCILLARY_COLOUR_SCIENCE_PACKAGES: Dict[str, str] = {}
"""
Ancillary *colour-science.org* packages to describe.

ANCILLARY_COLOUR_SCIENCE_PACKAGES
"""

ANCILLARY_RUNTIME_PACKAGES: Dict[str, str] = {}
"""
Ancillary runtime packages to describe.

ANCILLARY_RUNTIME_PACKAGES
"""

ANCILLARY_DEVELOPMENT_PACKAGES: Dict[str, str] = {}
"""
Ancillary development packages to describe.

ANCILLARY_DEVELOPMENT_PACKAGES
"""

ANCILLARY_EXTRAS_PACKAGES: Dict[str, str] = {}
"""
Ancillary extras packages to describe.

ANCILLARY_EXTRAS_PACKAGES
"""


def describe_environment(
    runtime_packages: Boolean = True,
    development_packages: Boolean = False,
    extras_packages: Boolean = False,
    print_environment: Boolean = True,
    **kwargs: Any,
) -> defaultdict:
    """
    Describe *Colour* running environment, i.e. interpreter, runtime and
    development packages.

    Parameters
    ----------
    runtime_packages
        Whether to return the runtime packages versions.
    development_packages
        Whether to return the development packages versions.
    extras_packages
        Whether to return the extras packages versions.
    print_environment
        Whether to print the environment.

    Other Parameters
    ----------------
    padding
        {:func:`colour.utilities.message_box`},
        Padding on each side of the message.
    print_callable
        {:func:`colour.utilities.message_box`},
        Callable used to print the message box.
    width
        {:func:`colour.utilities.message_box`},
        Message box width.

    Returns
    -------
    :class:`collections.defaultdict`
        Environment.

    Examples
    --------
    >>> environment = describe_environment(width=75)  # doctest: +SKIP
    ===========================================================================
    *                                                                         *
    *   Interpreter :                                                         *
    *       python : 3.8.6 (default, Nov 20 2020, 18:29:40)                   *
    *                [Clang 12.0.0 (clang-1200.0.32.27)]                      *
    *                                                                         *
    *   colour-science.org :                                                  *
    *       colour : v0.3.16-3-gd8bac475                                      *
    *                                                                         *
    *   Runtime :                                                             *
    *       imageio : 2.9.0                                                   *
    *       matplotlib : 3.3.3                                                *
    *       networkx : 2.5                                                    *
    *       numpy : 1.19.4                                                    *
    *       pandas : 0.25.3                                                   *
    *       pygraphviz : 1.6                                                  *
    *       scipy : 1.5.4                                                     *
    *       tqdm : 4.54.0                                                     *
    *                                                                         *
    ===========================================================================
    >>> environment = describe_environment(True, True, True, width=75)
    ... # doctest: +SKIP
    ===========================================================================
    *                                                                         *
    *   Interpreter :                                                         *
    *       python : 3.8.6 (default, Nov 20 2020, 18:29:40)                   *
    *                [Clang 12.0.0 (clang-1200.0.32.27)]                      *
    *                                                                         *
    *   colour-science.org :                                                  *
    *       colour : v0.3.16-3-gd8bac475                                      *
    *                                                                         *
    *   Runtime :                                                             *
    *       imageio : 2.9.0                                                   *
    *       matplotlib : 3.3.3                                                *
    *       networkx : 2.5                                                    *
    *       numpy : 1.19.4                                                    *
    *       pandas : 0.25.3                                                   *
    *       pygraphviz : 1.6                                                  *
    *       scipy : 1.5.4                                                     *
    *       tqdm : 4.54.0                                                     *
    *                                                                         *
    *   Development :                                                         *
    *       biblib-simple : 0.1.1                                             *
    *       coverage : 5.3                                                    *
    *       coveralls : 2.2.0                                                 *
    *       flake8 : 3.8.4                                                    *
    *       invoke : 1.4.1                                                    *
    *       jupyter : 1.0.0                                                   *
    *       mock : 4.0.2                                                      *
    *       nose : 1.3.7                                                      *
    *       pre-commit : 2.1.1                                                *
    *       pytest : 6.1.2                                                    *
    *       restructuredtext-lint : 1.3.2                                     *
    *       sphinx : 3.1.2                                                    *
    *       sphinx_rtd_theme : 0.5.0                                          *
    *       sphinxcontrib-bibtex : 1.0.0                                      *
    *       toml : 0.10.2                                                     *
    *       twine : 3.2.0                                                     *
    *       yapf : 0.23.0                                                     *
    *                                                                         *
    *   Extras :                                                              *
    *       ipywidgets : 7.5.1                                                *
    *       notebook : 6.1.5                                                  *
    *                                                                         *
    ===========================================================================
    """

    environment: defaultdict = defaultdict(dict)

    environment["Interpreter"]["python"] = sys.version

    import subprocess  # nosec

    import colour

    # TODO: Implement support for "pyproject.toml" file whenever "TOML" is
    # supported in the standard library.

    # NOTE: A few clauses are not reached and a few packages are not available
    # during continuous integration and are thus ignored for coverage.
    try:  # pragma: no cover
        output = subprocess.check_output(  # nosec
            ["git", "describe"],
            cwd=colour.__path__[0],
            stderr=subprocess.STDOUT,
        ).strip()
        version = output.decode("utf-8")
    except Exception:  # pragma: no cover
        version = colour.__version__

    environment["colour-science.org"]["colour"] = version
    environment["colour-science.org"].update(ANCILLARY_COLOUR_SCIENCE_PACKAGES)

    if runtime_packages:
        for package in [
            "imageio",
            "matplotlib",
            "networkx",
            "numpy",
            "pandas",
            "pygraphviz",
            "PyOpenColorIO",
            "scipy",
            "sklearn",
            "tqdm",
            "trimesh",
        ]:
            try:
                namespace = __import__(package)
                environment["Runtime"][package] = namespace.__version__
            except ImportError:
                continue

        # OpenImageIO
        try:  # pragma: no cover
            namespace = __import__("OpenImageIO")
            environment["Runtime"]["OpenImageIO"] = namespace.VERSION_STRING
        except ImportError:  # pragma: no cover
            pass

        environment["Runtime"].update(ANCILLARY_RUNTIME_PACKAGES)

    def _get_package_version(package: str, mapping: Mapping) -> str:
        """Return given package version."""

        namespace = __import__(package)

        if package in mapping:
            import pkg_resources

            distributions = [
                distribution for distribution in pkg_resources.working_set
            ]

            for distribution in distributions:
                if distribution.project_name == mapping[package]:
                    return distribution.version

        return namespace.__version__

    if development_packages:
        mapping = {
            "biblib.bib": "biblib-simple",
            "pre_commit": "pre-commit",
            "restructuredtext_lint": "restructuredtext-lint",
            "sphinxcontrib.bibtex": "sphinxcontrib-bibtex",
        }
        for package in [
            "biblib.bib",
            "coverage",
            "coveralls",
            "flake8",
            "invoke",
            "jupyter",
            "mock",
            "nose",
            "pre_commit",
            "pytest",
            "restructuredtext_lint",
            "sphinx",
            "sphinx_rtd_theme",
            "sphinxcontrib.bibtex",
            "toml",
            "twine",
            "yapf",
        ]:
            try:
                version = _get_package_version(package, mapping)
                package = mapping.get(package, package)

                environment["Development"][package] = version
            except Exception:  # pragma: no cover
                # pylint: disable=B112
                continue

        environment["Development"].update(ANCILLARY_DEVELOPMENT_PACKAGES)

    if extras_packages:
        mapping = {}
        for package in ["ipywidgets", "notebook"]:
            try:
                version = _get_package_version(package, mapping)
                package = mapping.get(package, package)

                environment["Extras"][package] = version
            except Exception:  # pragma: no cover
                # pylint: disable=B112
                continue

        environment["Extras"].update(ANCILLARY_EXTRAS_PACKAGES)

    if print_environment:
        message = ""
        for category in (
            "Interpreter",
            "colour-science.org",
            "Runtime",
            "Development",
            "Extras",
        ):
            elements = environment.get(category)
            if not elements:
                continue

            message += f"{category} :\n"
            for key, value in elements.items():
                lines = value.split("\n")
                message += f"    {key} : {lines.pop(0)}\n"
                indentation = len(f"    {key} : ")
                for line in lines:
                    message += f"{' ' * indentation}{line}\n"

            message += "\n"

        message_box(message.strip(), **kwargs)

    return environment



"""Defines the unit tests for the :mod:`colour.models.oklab` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.models import XYZ_to_Oklab, Oklab_to_XYZ
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestXYZ_to_Oklab",
    "TestOklab_to_XYZ",
]


class TestXYZ_to_Oklab(unittest.TestCase):
    """
    Define :func:`colour.models.oklab.TestXYZ_to_Oklab` definition unit
    tests methods.
    """

    def test_XYZ_to_Oklab(self):
        """Test :func:`colour.models.oklab.XYZ_to_Oklab` definition."""

        np.testing.assert_almost_equal(
            XYZ_to_Oklab(np.array([0.20654008, 0.12197225, 0.05136952])),
            np.array([0.51634019, 0.15469500, 0.06289579]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_Oklab(np.array([0.14222010, 0.23042768, 0.10495772])),
            np.array([0.59910746, -0.11139207, 0.07508465]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            XYZ_to_Oklab(np.array([0.96907232, 1.00000000, 1.12179215])),
            np.array([1.00121561, 0.00899591, -0.00535107]),
            decimal=7,
        )

    def test_n_dimensional_XYZ_to_Oklab(self):
        """
        Test :func:`colour.models.oklab.XYZ_to_Oklab` definition
        n-dimensional support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        Oklab = XYZ_to_Oklab(XYZ)

        XYZ = np.tile(XYZ, (6, 1))
        Oklab = np.tile(Oklab, (6, 1))
        np.testing.assert_almost_equal(XYZ_to_Oklab(XYZ), Oklab, decimal=7)

        XYZ = np.reshape(XYZ, (2, 3, 3))
        Oklab = np.reshape(Oklab, (2, 3, 3))
        np.testing.assert_almost_equal(XYZ_to_Oklab(XYZ), Oklab, decimal=7)

    def test_domain_range_scale_XYZ_to_Oklab(self):
        """
        Test :func:`colour.models.oklab.XYZ_to_Oklab` definition domain and
        range scale support.
        """

        XYZ = np.array([0.20654008, 0.12197225, 0.05136952])
        Oklab = XYZ_to_Oklab(XYZ)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    XYZ_to_Oklab(XYZ * factor), Oklab * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_XYZ_to_Oklab(self):
        """
        Test :func:`colour.models.oklab.XYZ_to_Oklab` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            XYZ = np.array(case)
            XYZ_to_Oklab(XYZ)


class TestOklab_to_XYZ(unittest.TestCase):
    """
    Define :func:`colour.models.oklab.Oklab_to_XYZ` definition unit tests
    methods.
    """

    def test_Oklab_to_XYZ(self):
        """Test :func:`colour.models.oklab.Oklab_to_XYZ` definition."""

        np.testing.assert_allclose(
            Oklab_to_XYZ(np.array([0.51634019, 0.15469500, 0.06289579])),
            np.array([0.20654008, 0.12197225, 0.05136952]),
            rtol=0.000001,
            atol=0.000001,
        )

        np.testing.assert_allclose(
            Oklab_to_XYZ(np.array([0.59910746, -0.11139207, 0.07508465])),
            np.array([0.14222010, 0.23042768, 0.10495772]),
            rtol=0.000001,
            atol=0.000001,
        )

        np.testing.assert_allclose(
            Oklab_to_XYZ(np.array([1.00121561, 0.00899591, -0.00535107])),
            np.array([0.96907232, 1.00000000, 1.12179215]),
            rtol=0.000001,
            atol=0.000001,
        )

    def test_n_dimensional_Oklab_to_XYZ(self):
        """
        Test :func:`colour.models.oklab.Oklab_to_XYZ` definition
        n-dimensional support.
        """

        Oklab = np.array([0.51634019, 0.15469500, 0.06289579])
        XYZ = Oklab_to_XYZ(Oklab)

        Oklab = np.tile(Oklab, (6, 1))
        XYZ = np.tile(XYZ, (6, 1))
        np.testing.assert_allclose(
            Oklab_to_XYZ(Oklab), XYZ, rtol=0.000001, atol=0.000001
        )

        Oklab = np.reshape(Oklab, (2, 3, 3))
        XYZ = np.reshape(XYZ, (2, 3, 3))
        np.testing.assert_allclose(
            Oklab_to_XYZ(Oklab), XYZ, rtol=0.000001, atol=0.000001
        )

    def test_domain_range_scale_Oklab_to_XYZ(self):
        """
        Test :func:`colour.models.oklab.Oklab_to_XYZ` definition domain and
        range scale support.
        """

        Oklab = np.array([0.51634019, 0.15469500, 0.06289579])
        XYZ = Oklab_to_XYZ(Oklab)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_allclose(
                    Oklab_to_XYZ(Oklab * factor),
                    XYZ * factor,
                    rtol=0.000001,
                    atol=0.000001,
                )

    @ignore_numpy_errors
    def test_nan_Oklab_to_XYZ(self):
        """
        Test :func:`colour.models.oklab.Oklab_to_XYZ` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            Oklab = np.array(case)
            Oklab_to_XYZ(Oklab)


if __name__ == "__main__":
    unittest.main()



"""
Fujifilm F-Gamut Colourspace
============================

Defines the *Fujifilm F-Gamut* colourspace:

-   :attr:`colour.models.RGB_COLOURSPACE_F_GAMUT`.

References
----------
-   :cite:`Fujifilm2016` : Fujifilm. (2016). F-Log Data Sheet Ver.1.0 (pp.
    1-4). https://www.fujifilm.com/support/digital_cameras/software/lut/pdf/\
F-Log_DataSheet_E_Ver.1.0.pdf
"""

from __future__ import annotations

import numpy as np

from colour.colorimetry import CCS_ILLUMINANTS
from colour.hints import NDArray
from colour.models.rgb import (
    RGB_Colourspace,
    log_encoding_FLog,
    normalised_primary_matrix,
    log_decoding_FLog,
)

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - http://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "PRIMARIES_F_GAMUT",
    "WHITEPOINT_NAME_F_GAMUT",
    "CCS_WHITEPOINT_F_GAMUT",
    "MATRIX_F_GAMUT_TO_XYZ",
    "MATRIX_XYZ_TO_F_GAMUT",
    "RGB_COLOURSPACE_F_GAMUT",
]

PRIMARIES_F_GAMUT: NDArray = np.array(
    [
        [0.70800, 0.29200],
        [0.17000, 0.79700],
        [0.13100, 0.04600],
    ]
)
"""*Fujifilm F-Gamut* colourspace primaries."""

WHITEPOINT_NAME_F_GAMUT: str = "D65"
"""*Fujifilm F-Gamut* colourspace whitepoint name."""

CCS_WHITEPOINT_F_GAMUT: NDArray = CCS_ILLUMINANTS[
    "CIE 1931 2 Degree Standard Observer"
][WHITEPOINT_NAME_F_GAMUT]
"""*Fujifilm F-Gamut* colourspace whitepoint chromaticity coordinates."""

MATRIX_F_GAMUT_TO_XYZ: NDArray = normalised_primary_matrix(
    PRIMARIES_F_GAMUT, CCS_WHITEPOINT_F_GAMUT
)
"""*Fujifilm F-Gamut* colourspace to *CIE XYZ* tristimulus values matrix."""

MATRIX_XYZ_TO_F_GAMUT: NDArray = np.linalg.inv(MATRIX_F_GAMUT_TO_XYZ)
"""*CIE XYZ* tristimulus values to *Fujifilm F-Gamut* colourspace matrix."""

RGB_COLOURSPACE_F_GAMUT: RGB_Colourspace = RGB_Colourspace(
    "F-Gamut",
    PRIMARIES_F_GAMUT,
    CCS_WHITEPOINT_F_GAMUT,
    WHITEPOINT_NAME_F_GAMUT,
    MATRIX_F_GAMUT_TO_XYZ,
    MATRIX_XYZ_TO_F_GAMUT,
    log_encoding_FLog,
    log_decoding_FLog,
)
RGB_COLOURSPACE_F_GAMUT.__doc__ = """
*Fujifilm F-Gamut* colourspace.

References
----------
:cite:`Fujifilm2016`
"""



"""
Defines the unit tests for the :mod:`colour.models.rgb.transfer_functions.\
panasonic_vlog` module.
"""

import numpy as np
import unittest

from colour.models.rgb.transfer_functions import (
    log_encoding_VLog,
    log_decoding_VLog,
)
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestLogEncoding_VLog",
    "TestLogDecoding_VLog",
]


class TestLogEncoding_VLog(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.panasonic_vlog.\
log_encoding_VLog` definition unit tests methods.
    """

    def test_log_encoding_VLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.panasonic_vlog.\
log_encoding_VLog` definition.
        """

        self.assertAlmostEqual(log_encoding_VLog(0.0), 0.125, places=7)

        self.assertAlmostEqual(
            log_encoding_VLog(0.18), 0.423311448760136, places=7
        )

        self.assertAlmostEqual(
            log_encoding_VLog(0.18, 12), 0.423311448760136, places=7
        )

        self.assertAlmostEqual(
            log_encoding_VLog(0.18, 10, False), 0.421287228403675, places=7
        )

        self.assertAlmostEqual(
            log_encoding_VLog(0.18, 10, False, False),
            0.409009628526078,
            places=7,
        )

        self.assertAlmostEqual(
            log_encoding_VLog(1.0), 0.599117700158146, places=7
        )

    def test_n_dimensional_log_encoding_VLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.panasonic_vlog.\
log_encoding_VLog` definition n-dimensional arrays support.
        """

        L_in = 0.18
        V_out = log_encoding_VLog(L_in)

        L_in = np.tile(L_in, 6)
        V_out = np.tile(V_out, 6)
        np.testing.assert_almost_equal(
            log_encoding_VLog(L_in), V_out, decimal=7
        )

        L_in = np.reshape(L_in, (2, 3))
        V_out = np.reshape(V_out, (2, 3))
        np.testing.assert_almost_equal(
            log_encoding_VLog(L_in), V_out, decimal=7
        )

        L_in = np.reshape(L_in, (2, 3, 1))
        V_out = np.reshape(V_out, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_encoding_VLog(L_in), V_out, decimal=7
        )

    def test_domain_range_scale_log_encoding_VLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.panasonic_vlog.\
log_encoding_VLog` definition domain and range scale support.
        """

        L_in = 0.18
        V_out = log_encoding_VLog(L_in)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_encoding_VLog(L_in * factor), V_out * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_log_encoding_VLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.panasonic_vlog.\
log_encoding_VLog` definition nan support.
        """

        log_encoding_VLog(np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]))


class TestLogDecoding_VLog(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.transfer_functions.panasonic_vlog.\
log_decoding_VLog` definition unit tests methods.
    """

    def test_log_decoding_VLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.panasonic_vlog.\
log_decoding_VLog` definition.
        """

        self.assertAlmostEqual(log_decoding_VLog(0.125), 0.0, places=7)

        self.assertAlmostEqual(
            log_decoding_VLog(0.423311448760136), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_VLog(0.423311448760136, 12), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_VLog(0.421287228403675, 10, False), 0.18, places=7
        )

        self.assertAlmostEqual(
            log_decoding_VLog(0.409009628526078, 10, False, False),
            0.18,
            places=7,
        )

        self.assertAlmostEqual(
            log_decoding_VLog(0.599117700158146), 1.0, places=7
        )

    def test_n_dimensional_log_decoding_VLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.panasonic_vlog.\
log_decoding_VLog` definition n-dimensional arrays support.
        """

        V_out = 0.423311448760136
        L_in = log_decoding_VLog(V_out)

        V_out = np.tile(V_out, 6)
        L_in = np.tile(L_in, 6)
        np.testing.assert_almost_equal(
            log_decoding_VLog(V_out), L_in, decimal=7
        )

        V_out = np.reshape(V_out, (2, 3))
        L_in = np.reshape(L_in, (2, 3))
        np.testing.assert_almost_equal(
            log_decoding_VLog(V_out), L_in, decimal=7
        )

        V_out = np.reshape(V_out, (2, 3, 1))
        L_in = np.reshape(L_in, (2, 3, 1))
        np.testing.assert_almost_equal(
            log_decoding_VLog(V_out), L_in, decimal=7
        )

    def test_domain_range_scale_log_decoding_VLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.panasonic_vlog.\
log_decoding_VLog` definition domain and range scale support.
        """

        V_out = 0.423311448760136
        L_in = log_decoding_VLog(V_out)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    log_decoding_VLog(V_out * factor), L_in * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_log_decoding_VLog(self):
        """
        Test :func:`colour.models.rgb.transfer_functions.panasonic_vlog.\
log_decoding_VLog` definition nan support.
        """

        log_decoding_VLog(np.array([-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]))


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.continuous.signal` module."""

import numpy as np
import unittest
import re
import textwrap

from colour.algebra import (
    CubicSplineInterpolator,
    Extrapolator,
    KernelInterpolator,
)
from colour.constants import DEFAULT_FLOAT_DTYPE
from colour.continuous import Signal
from colour.utilities import ColourRuntimeWarning, attest, is_pandas_installed

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestSignal",
]


class TestSignal(unittest.TestCase):
    """Define :class:`colour.continuous.signal.Signal` class unit tests methods."""

    def setUp(self):
        """Initialise the common tests attributes."""

        self._range = np.linspace(10, 100, 10)
        self._domain = np.arange(100, 1100, 100)

        self._signal = Signal(self._range)

    def test_required_attributes(self):
        """Test the presence of required attributes."""

        required_attributes = (
            "dtype",
            "domain",
            "range",
            "interpolator",
            "interpolator_kwargs",
            "extrapolator",
            "extrapolator_kwargs",
            "function",
        )

        for attribute in required_attributes:
            self.assertIn(attribute, dir(Signal))

    def test_required_methods(self):
        """Test the presence of required methods."""

        required_methods = (
            "__init__",
            "__str__",
            "__repr__",
            "__hash__",
            "__getitem__",
            "__setitem__",
            "__contains__",
            "__eq__",
            "__ne__",
            "arithmetical_operation",
            "signal_unpack_data",
            "fill_nan",
            "domain_distance",
            "to_series",
        )

        for method in required_methods:
            self.assertIn(method, dir(Signal))

    def test_dtype(self):
        """Test :func:`colour.continuous.signal.Signal.dtype` property."""

        self.assertEqual(self._signal.dtype, DEFAULT_FLOAT_DTYPE)

        signal = self._signal.copy()
        signal.dtype = np.float32
        self.assertEqual(signal.dtype, np.float32)

    def test_domain(self):
        """Test :func:`colour.continuous.signal.Signal.domain` property."""

        signal = self._signal.copy()

        np.testing.assert_almost_equal(
            signal[np.array([0, 1, 2])],
            np.array([10.0, 20.0, 30.0]),
            decimal=7,
        )

        signal.domain = np.arange(0, 10, 1) * 10

        np.testing.assert_array_equal(signal.domain, np.arange(0, 10, 1) * 10)

        np.testing.assert_almost_equal(
            signal[np.array([0, 1, 2]) * 10],
            np.array([10.0, 20.0, 30.0]),
            decimal=7,
        )

        domain = np.linspace(0, 1, 10)
        domain[0] = -np.inf

        def assert_warns():
            """Help to test the runtime warning."""

            signal.domain = domain

        self.assertWarns(ColourRuntimeWarning, assert_warns)

    def test_range(self):
        """Test :func:`colour.continuous.signal.Signal.range` property."""

        signal = self._signal.copy()

        np.testing.assert_almost_equal(
            signal[np.array([0, 1, 2])],
            np.array([10.0, 20.0, 30.0]),
            decimal=7,
        )

        signal.range = self._range * 10

        np.testing.assert_array_equal(signal.range, self._range * 10)

        np.testing.assert_almost_equal(
            signal[np.array([0, 1, 2])],
            np.array([10.0, 20.0, 30.0]) * 10,
            decimal=7,
        )

    def test_interpolator(self):
        """Test :func:`colour.continuous.signal.Signal.interpolator` property."""

        signal = self._signal.copy()

        np.testing.assert_almost_equal(
            signal[np.linspace(0, 5, 5)],
            np.array(
                [
                    10.00000000,
                    22.83489024,
                    34.80044921,
                    47.55353925,
                    60.00000000,
                ]
            ),
            decimal=7,
        )

        signal.interpolator = CubicSplineInterpolator

        np.testing.assert_almost_equal(
            signal[np.linspace(0, 5, 5)],
            np.array([10.0, 22.5, 35.0, 47.5, 60.0]),
            decimal=7,
        )

    def test_interpolator_kwargs(self):
        """
        Test :func:`colour.continuous.signal.Signal.interpolator_kwargs`
        property.
        """

        signal = self._signal.copy()

        np.testing.assert_almost_equal(
            signal[np.linspace(0, 5, 5)],
            np.array(
                [
                    10.00000000,
                    22.83489024,
                    34.80044921,
                    47.55353925,
                    60.00000000,
                ]
            ),
            decimal=7,
        )

        signal.interpolator_kwargs = {"window": 1, "kernel_kwargs": {"a": 1}}

        np.testing.assert_almost_equal(
            signal[np.linspace(0, 5, 5)],
            np.array(
                [
                    10.00000000,
                    18.91328761,
                    28.36993142,
                    44.13100443,
                    60.00000000,
                ]
            ),
            decimal=7,
        )

    def test_extrapolator(self):
        """Test :func:`colour.continuous.signal.Signal.extrapolator` property."""

        self.assertIsInstance(self._signal.extrapolator(), Extrapolator)

    def test_extrapolator_kwargs(self):
        """
        Test :func:`colour.continuous.signal.Signal.extrapolator_kwargs`
        property.
        """

        signal = self._signal.copy()

        attest(np.all(np.isnan(signal[np.array([-1000, 1000])])))

        signal.extrapolator_kwargs = {
            "method": "Linear",
        }

        np.testing.assert_almost_equal(
            signal[np.array([-1000, 1000])],
            np.array([-9990.0, 10010.0]),
            decimal=7,
        )

    def test_function(self):
        """Test :func:`colour.continuous.signal.Signal.function` property."""

        attest(hasattr(self._signal.function, "__call__"))

    def test_raise_exception_function(self):
        """
        Test :func:`colour.continuous.signal.Signal.function` property raised
        exception.
        """

        self.assertRaises(ValueError, Signal().function, 0)

    def test__init__(self):
        """Test :func:`colour.continuous.signal.Signal.__init__` method."""

        signal = Signal(self._range)
        np.testing.assert_array_equal(signal.domain, np.arange(0, 10, 1))
        np.testing.assert_array_equal(signal.range, self._range)

        signal = Signal(self._range, self._domain)
        np.testing.assert_array_equal(signal.domain, self._domain)
        np.testing.assert_array_equal(signal.range, self._range)

        signal = Signal(dict(zip(self._domain, self._range)))
        np.testing.assert_array_equal(signal.domain, self._domain)
        np.testing.assert_array_equal(signal.range, self._range)

        signal = Signal(signal)
        np.testing.assert_array_equal(signal.domain, self._domain)
        np.testing.assert_array_equal(signal.range, self._range)

        if is_pandas_installed():
            from pandas import Series

            signal = Signal(Series(dict(zip(self._domain, self._range))))
            np.testing.assert_array_equal(signal.domain, self._domain)
            np.testing.assert_array_equal(signal.range, self._range)

    def test__hash__(self):
        """Test :func:`colour.continuous.signal.Signal.__hash__` method."""

        self.assertIsInstance(hash(self._signal), int)

    def test__str__(self):
        """Test :func:`colour.continuous.signal.Signal.__str__` method."""

        self.assertEqual(
            str(self._signal),
            textwrap.dedent(
                """
                [[   0.   10.]
                 [   1.   20.]
                 [   2.   30.]
                 [   3.   40.]
                 [   4.   50.]
                 [   5.   60.]
                 [   6.   70.]
                 [   7.   80.]
                 [   8.   90.]
                 [   9.  100.]]"""
            )[1:],
        )

        self.assertIsInstance(str(Signal()), str)

    def test__repr__(self):
        """Test :func:`colour.continuous.signal.Signal.__repr__` method."""

        self.assertEqual(
            re.sub(
                r"extrapolator_kwargs={.*}",
                "extrapolator_kwargs={...}",
                repr(self._signal),
            ),
            textwrap.dedent(
                """
                Signal([[   0.,   10.],
                        [   1.,   20.],
                        [   2.,   30.],
                        [   3.,   40.],
                        [   4.,   50.],
                        [   5.,   60.],
                        [   6.,   70.],
                        [   7.,   80.],
                        [   8.,   90.],
                        [   9.,  100.]],
                       interpolator=KernelInterpolator,
                       interpolator_kwargs={},
                       extrapolator=Extrapolator,
                       extrapolator_kwargs={...})"""
            )[1:],
        )

        self.assertIsInstance(repr(Signal()), str)

    def test__getitem__(self):
        """Test :func:`colour.continuous.signal.Signal.__getitem__` method."""

        self.assertEqual(self._signal[0], 10.0)

        np.testing.assert_almost_equal(
            self._signal[np.array([0, 1, 2])],
            np.array([10.0, 20.0, 30.0]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._signal[np.linspace(0, 5, 5)],
            np.array(
                [
                    10.00000000,
                    22.83489024,
                    34.80044921,
                    47.55353925,
                    60.00000000,
                ]
            ),
            decimal=7,
        )

        attest(np.all(np.isnan(self._signal[np.array([-1000, 1000])])))

        signal = self._signal.copy()
        signal.extrapolator_kwargs = {
            "method": "Linear",
        }
        np.testing.assert_array_equal(
            signal[np.array([-1000, 1000])], np.array([-9990.0, 10010.0])
        )

        signal.extrapolator_kwargs = {
            "method": "Constant",
            "left": 0,
            "right": 1,
        }
        np.testing.assert_array_equal(
            signal[np.array([-1000, 1000])], np.array([0.0, 1.0])
        )

    def test__setitem__(self):
        """Test :func:`colour.continuous.signal.Signal.__setitem__` method."""

        signal = self._signal.copy()

        signal[0] = 20
        np.testing.assert_almost_equal(
            signal.range,
            np.array(
                [20.0, 20.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0]
            ),
        )

        signal[np.array([0, 1, 2])] = 30
        np.testing.assert_almost_equal(
            signal.range,
            np.array(
                [30.0, 30.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0]
            ),
            decimal=7,
        )

        signal[0:3] = 40
        np.testing.assert_almost_equal(
            signal.range,
            np.array(
                [40.0, 40.0, 40.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0]
            ),
            decimal=7,
        )

        signal[np.linspace(0, 5, 5)] = 50
        np.testing.assert_almost_equal(
            signal.domain,
            np.array(
                [
                    0.00,
                    1.00,
                    1.25,
                    2.00,
                    2.50,
                    3.00,
                    3.75,
                    4.00,
                    5.00,
                    6.00,
                    7.00,
                    8.00,
                    9.00,
                ]
            ),
            decimal=7,
        )
        np.testing.assert_almost_equal(
            signal.range,
            np.array(
                [
                    50.0,
                    40.0,
                    50.0,
                    40.0,
                    50.0,
                    40.0,
                    50.0,
                    50.0,
                    50.0,
                    70.0,
                    80.0,
                    90.0,
                    100.0,
                ]
            ),
            decimal=7,
        )

        signal[np.array([0, 1, 2])] = np.array([10, 20, 30])
        np.testing.assert_almost_equal(
            signal.range,
            np.array(
                [
                    10.0,
                    20.0,
                    50.0,
                    30.0,
                    50.0,
                    40.0,
                    50.0,
                    50.0,
                    50.0,
                    70.0,
                    80.0,
                    90.0,
                    100.0,
                ]
            ),
            decimal=7,
        )

    def test__contains__(self):
        """Test :func:`colour.continuous.signal.Signal.__contains__` method."""

        self.assertIn(0, self._signal)
        self.assertIn(0.5, self._signal)
        self.assertNotIn(1000, self._signal)

    def test__len__(self):
        """Test :func:`colour.continuous.signal.Signal.__len__` method."""

        self.assertEqual(len(self._signal), 10)

    def test__eq__(self):
        """Test :func:`colour.continuous.signal.Signal.__eq__` method."""

        signal_1 = self._signal.copy()
        signal_2 = self._signal.copy()

        self.assertEqual(signal_1, signal_2)

    def test__ne__(self):
        """Test :func:`colour.continuous.signal.Signal.__ne__` method."""

        signal_1 = self._signal.copy()
        signal_2 = self._signal.copy()

        signal_2[0] = 20
        self.assertNotEqual(signal_1, signal_2)

        signal_2[0] = 10
        self.assertEqual(signal_1, signal_2)

        signal_2.interpolator = CubicSplineInterpolator
        self.assertNotEqual(signal_1, signal_2)

        signal_2.interpolator = KernelInterpolator
        self.assertEqual(signal_1, signal_2)

        signal_2.interpolator_kwargs = {"window": 1}
        self.assertNotEqual(signal_1, signal_2)

        signal_2.interpolator_kwargs = {}
        self.assertEqual(signal_1, signal_2)

        class NotExtrapolator(Extrapolator):
            """Not :class:`Extrapolator` class."""

            pass

        signal_2.extrapolator = NotExtrapolator
        self.assertNotEqual(signal_1, signal_2)

        signal_2.extrapolator = Extrapolator
        self.assertEqual(signal_1, signal_2)

        signal_2.extrapolator_kwargs = {}
        self.assertNotEqual(signal_1, signal_2)

        signal_2.extrapolator_kwargs = {
            "method": "Constant",
            "left": np.nan,
            "right": np.nan,
        }
        self.assertEqual(signal_1, signal_2)

    def test_arithmetical_operation(self):
        """
        Test :meth:`colour.continuous.signal.Signal.arithmetical_operation`
        method.
        """

        np.testing.assert_almost_equal(
            self._signal.arithmetical_operation(10, "+", False).range,
            self._range + 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._signal.arithmetical_operation(10, "-", False).range,
            self._range - 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._signal.arithmetical_operation(10, "*", False).range,
            self._range * 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._signal.arithmetical_operation(10, "/", False).range,
            self._range / 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            self._signal.arithmetical_operation(10, "**", False).range,
            self._range**10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            (self._signal + 10).range, self._range + 10, decimal=7
        )

        np.testing.assert_almost_equal(
            (self._signal - 10).range, self._range - 10, decimal=7
        )

        np.testing.assert_almost_equal(
            (self._signal * 10).range, self._range * 10, decimal=7
        )

        np.testing.assert_almost_equal(
            (self._signal / 10).range, self._range / 10, decimal=7
        )

        np.testing.assert_almost_equal(
            (self._signal**10).range, self._range**10, decimal=7
        )

        signal = self._signal.copy()

        np.testing.assert_almost_equal(
            signal.arithmetical_operation(10, "+", True).range,
            self._range + 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            signal.arithmetical_operation(10, "-", True).range,
            self._range,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            signal.arithmetical_operation(10, "*", True).range,
            self._range * 10,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            signal.arithmetical_operation(10, "/", True).range,
            self._range,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            signal.arithmetical_operation(10, "**", True).range,
            self._range**10,
            decimal=7,
        )

        signal = self._signal.copy()

        np.testing.assert_almost_equal(
            signal.arithmetical_operation(self._range, "+", False).range,
            signal.range + self._range,
            decimal=7,
        )

        np.testing.assert_almost_equal(
            signal.arithmetical_operation(signal, "+", False).range,
            signal.range + signal._range,
            decimal=7,
        )

    def test_is_uniform(self):
        """Test :func:`colour.continuous.signal.Signal.is_uniform` method."""

        self.assertTrue(self._signal.is_uniform())

        signal = self._signal.copy()
        signal[0.5] = 1.0
        self.assertFalse(signal.is_uniform())

    def test_copy(self):
        """Test :func:`colour.continuous.signal.Signal.copy` method."""

        self.assertIsNot(self._signal, self._signal.copy())
        self.assertEqual(self._signal, self._signal.copy())

    def test_signal_unpack_data(self):
        """
        Test :meth:`colour.continuous.signal.Signal.signal_unpack_data`
        method.
        """

        domain, range_ = Signal.signal_unpack_data(self._range)
        np.testing.assert_array_equal(range_, self._range)
        np.testing.assert_array_equal(domain, np.arange(0, 10, 1))

        domain, range_ = Signal.signal_unpack_data(self._range, self._domain)
        np.testing.assert_array_equal(range_, self._range)
        np.testing.assert_array_equal(domain, self._domain)

        domain, range_ = Signal.signal_unpack_data(
            self._range, dict(zip(self._domain, self._range)).keys()
        )
        np.testing.assert_array_equal(domain, self._domain)

        domain, range_ = Signal.signal_unpack_data(
            dict(zip(self._domain, self._range))
        )
        np.testing.assert_array_equal(range_, self._range)
        np.testing.assert_array_equal(domain, self._domain)

        domain, range_ = Signal.signal_unpack_data(
            Signal(self._range, self._domain)
        )
        np.testing.assert_array_equal(range_, self._range)
        np.testing.assert_array_equal(domain, self._domain)

        if is_pandas_installed():
            from pandas import Series

            domain, range_ = Signal.signal_unpack_data(
                Series(dict(zip(self._domain, self._range)))
            )
            np.testing.assert_array_equal(range_, self._range)
            np.testing.assert_array_equal(domain, self._domain)

    def test_fill_nan(self):
        """Test :func:`colour.continuous.signal.Signal.fill_nan` method."""

        signal = self._signal.copy()

        signal[3:7] = np.nan

        np.testing.assert_almost_equal(
            signal.fill_nan().range,
            np.array(
                [10.0, 20.0, 30.0, 40.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0]
            ),
            decimal=7,
        )

        signal[3:7] = np.nan

        np.testing.assert_almost_equal(
            signal.fill_nan(method="Constant").range,
            np.array(
                [10.0, 20.0, 30.0, 0.0, 0.0, 0.0, 0.0, 80.0, 90.0, 100.0]
            ),
            decimal=7,
        )

    def test_domain_distance(self):
        """Test :func:`colour.continuous.signal.Signal.domain_distance` method."""

        self.assertAlmostEqual(
            self._signal.domain_distance(0.5), 0.5, places=7
        )

        np.testing.assert_almost_equal(
            self._signal.domain_distance(np.linspace(0, 9, 10) + 0.5),
            np.array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]),
            decimal=7,
        )

    def test_to_series(self):
        """Test :func:`colour.continuous.signal.Signal.to_series` method."""

        if is_pandas_installed():
            from pandas import Series

            self.assertEqual(
                Signal(self._range, self._domain).to_series().all(),
                Series(dict(zip(self._domain, self._range))).all(),
            )


if __name__ == "__main__":
    unittest.main()



"""Defines the unit tests for the :mod:`colour.models.rgb.cylindrical` module."""

import numpy as np
import unittest
from itertools import permutations

from colour.models.rgb.cylindrical import (
    RGB_to_HSV,
    HSV_to_RGB,
    RGB_to_HSL,
    HSL_to_RGB,
    RGB_to_HCL,
    HCL_to_RGB,
)
from colour.utilities import domain_range_scale, ignore_numpy_errors

__author__ = "Colour Developers"
__copyright__ = "Copyright 2013 Colour Developers"
__license__ = "New BSD License - https://opensource.org/licenses/BSD-3-Clause"
__maintainer__ = "Colour Developers"
__email__ = "colour-developers@colour-science.org"
__status__ = "Production"

__all__ = [
    "TestRGB_to_HSV",
    "TestHSV_to_RGB",
    "TestRGB_to_HSL",
    "TestHSL_to_RGB",
    "TestRGB_to_HCL",
    "TestHCL_to_RGB",
]


class TestRGB_to_HSV(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.cylindrical.RGB_to_HSV` definition unit
    tests methods.
    """

    def test_RGB_to_HSV(self):
        """Test :func:`colour.models.rgb.cylindrical.RGB_to_HSV` definition."""

        np.testing.assert_almost_equal(
            RGB_to_HSV(np.array([0.45620519, 0.03081071, 0.04091952])),
            np.array([0.99603944, 0.93246304, 0.45620519]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_HSV(np.array([0.00000000, 0.00000000, 0.00000000])),
            np.array([0.00000000, 0.00000000, 0.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_HSV(np.array([1.00000000, 1.00000000, 1.00000000])),
            np.array([0.00000000, 0.00000000, 1.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_HSV(np.array([0.00000000, 1.00000000, 1.00000000])),
            np.array([0.50000000, 1.00000000, 1.00000000]),
            decimal=7,
        )

    def test_n_dimensional_RGB_to_HSV(self):
        """
        Test :func:`colour.models.rgb.cylindrical.RGB_to_HSV` definition
        n-dimensional arrays support.
        """

        RGB = np.array([0.45620519, 0.03081071, 0.04091952])
        HSV = RGB_to_HSV(RGB)

        RGB = np.tile(RGB, (6, 1))
        HSV = np.tile(HSV, (6, 1))
        np.testing.assert_almost_equal(RGB_to_HSV(RGB), HSV, decimal=7)

        RGB = np.reshape(RGB, (2, 3, 3))
        HSV = np.reshape(HSV, (2, 3, 3))
        np.testing.assert_almost_equal(RGB_to_HSV(RGB), HSV, decimal=7)

    def test_domain_range_scale_RGB_to_HSV(self):
        """
        Test :func:`colour.models.rgb.cylindrical.RGB_to_HSV` definition
        domain and range scale support.
        """

        RGB = np.array([0.45620519, 0.03081071, 0.04091952])
        HSV = RGB_to_HSV(RGB)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    RGB_to_HSV(RGB * factor), HSV * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_RGB_to_HSV(self):
        """
        Test :func:`colour.models.rgb.cylindrical.RGB_to_HSV` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            RGB = np.array(case)
            RGB_to_HSV(RGB)


class TestHSV_to_RGB(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.cylindrical.HSV_to_RGB` definition unit
    tests methods.
    """

    def test_HSV_to_RGB(self):
        """Test :func:`colour.models.rgb.cylindrical.HSV_to_RGB` definition."""

        np.testing.assert_almost_equal(
            HSV_to_RGB(np.array([0.99603944, 0.93246304, 0.45620519])),
            np.array([0.45620519, 0.03081071, 0.04091952]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            HSV_to_RGB(np.array([0.00000000, 0.00000000, 0.00000000])),
            np.array([0.00000000, 0.00000000, 0.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            HSV_to_RGB(np.array([0.00000000, 0.00000000, 1.00000000])),
            np.array([1.00000000, 1.00000000, 1.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            HSV_to_RGB(np.array([0.50000000, 1.00000000, 1.00000000])),
            np.array([0.00000000, 1.00000000, 1.00000000]),
            decimal=7,
        )

    def test_n_dimensional_HSV_to_RGB(self):
        """
        Test :func:`colour.models.rgb.cylindrical.HSV_to_RGB` definition
        n-dimensional arrays support.
        """

        HSV = np.array([0.99603944, 0.93246304, 0.45620519])
        RGB = HSV_to_RGB(HSV)

        HSV = np.tile(HSV, (6, 1))
        RGB = np.tile(RGB, (6, 1))
        np.testing.assert_almost_equal(HSV_to_RGB(HSV), RGB, decimal=7)

        HSV = np.reshape(HSV, (2, 3, 3))
        RGB = np.reshape(RGB, (2, 3, 3))
        np.testing.assert_almost_equal(HSV_to_RGB(HSV), RGB, decimal=7)

    def test_domain_range_scale_HSV_to_RGB(self):
        """
        Test :func:`colour.models.rgb.cylindrical.HSV_to_RGB` definition
        domain and range scale support.
        """

        HSV = np.array([0.99603944, 0.93246304, 0.45620519])
        RGB = HSV_to_RGB(HSV)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    HSV_to_RGB(HSV * factor), RGB * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_HSV_to_RGB(self):
        """
        Test :func:`colour.models.rgb.cylindrical.HSV_to_RGB` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            HSV = np.array(case)
            HSV_to_RGB(HSV)


class TestRGB_to_HSL(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.cylindrical.RGB_to_HSL` definition unit
    tests methods.
    """

    def test_RGB_to_HSL(self):
        """Test :func:`colour.models.rgb.cylindrical.RGB_to_HSL` definition."""

        np.testing.assert_almost_equal(
            RGB_to_HSL(np.array([0.45620519, 0.03081071, 0.04091952])),
            np.array([0.99603944, 0.87347144, 0.24350795]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_HSL(np.array([0.00000000, 0.00000000, 0.00000000])),
            np.array([0.00000000, 0.00000000, 0.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_HSL(np.array([1.00000000, 1.00000000, 1.00000000])),
            np.array([0.00000000, 0.00000000, 1.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_HSL(np.array([1.00000000, 0.00000000, 0.00000000])),
            np.array([0.00000000, 1.00000000, 0.50000000]),
            decimal=7,
        )

    def test_n_dimensional_RGB_to_HSL(self):
        """
        Test :func:`colour.models.rgb.cylindrical.RGB_to_HSL` definition
        n-dimensional arrays support.
        """

        RGB = np.array([0.45620519, 0.03081071, 0.04091952])
        HSL = RGB_to_HSL(RGB)

        RGB = np.tile(RGB, (6, 1))
        HSL = np.tile(HSL, (6, 1))
        np.testing.assert_almost_equal(RGB_to_HSL(RGB), HSL, decimal=7)

        RGB = np.reshape(RGB, (2, 3, 3))
        HSL = np.reshape(HSL, (2, 3, 3))
        np.testing.assert_almost_equal(RGB_to_HSL(RGB), HSL, decimal=7)

    def test_domain_range_scale_RGB_to_HSL(self):
        """
        Test :func:`colour.models.rgb.cylindrical.RGB_to_HSL` definition
        domain and range scale support.
        """

        RGB = np.array([0.45620519, 0.03081071, 0.04091952])
        HSL = RGB_to_HSL(RGB)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    RGB_to_HSL(RGB * factor), HSL * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_RGB_to_HSL(self):
        """
        Test :func:`colour.models.rgb.cylindrical.RGB_to_HSL` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            RGB = np.array(case)
            RGB_to_HSL(RGB)


class TestHSL_to_RGB(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.cylindrical.HSL_to_RGB` definition unit
    tests methods.
    """

    def test_HSL_to_RGB(self):
        """Test :func:`colour.models.rgb.cylindrical.HSL_to_RGB` definition."""

        np.testing.assert_almost_equal(
            HSL_to_RGB(np.array([0.99603944, 0.87347144, 0.24350795])),
            np.array([0.45620519, 0.03081071, 0.04091952]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            HSL_to_RGB(np.array([0.00000000, 0.00000000, 0.00000000])),
            np.array([0.00000000, 0.00000000, 0.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            HSL_to_RGB(np.array([0.00000000, 0.00000000, 1.00000000])),
            np.array([1.00000000, 1.00000000, 1.00000000]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            HSL_to_RGB(np.array([0.00000000, 1.00000000, 0.50000000])),
            np.array([1.00000000, 0.00000000, 0.00000000]),
            decimal=7,
        )

    def test_n_dimensional_HSL_to_RGB(self):
        """
        Test :func:`colour.models.rgb.cylindrical.HSL_to_RGB` definition
        n-dimensional arrays support.
        """

        HSL = np.array([0.99603944, 0.87347144, 0.24350795])
        RGB = HSL_to_RGB(HSL)

        HSL = np.tile(HSL, (6, 1))
        RGB = np.tile(RGB, (6, 1))
        np.testing.assert_almost_equal(HSL_to_RGB(HSL), RGB, decimal=7)

        HSL = np.reshape(HSL, (2, 3, 3))
        RGB = np.reshape(RGB, (2, 3, 3))
        np.testing.assert_almost_equal(HSL_to_RGB(HSL), RGB, decimal=7)

    def test_domain_range_scale_HSL_to_RGB(self):
        """
        Test :func:`colour.models.rgb.cylindrical.HSL_to_RGB` definition
        domain and range scale support.
        """

        HSL = np.array([0.99603944, 0.87347144, 0.24350795])
        RGB = HSL_to_RGB(HSL)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    HSL_to_RGB(HSL * factor), RGB * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_HSL_to_RGB(self):
        """
        Test :func:`colour.models.rgb.cylindrical.HSL_to_RGB` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            HSL = np.array(case)
            HSL_to_RGB(HSL)


class TestRGB_to_HCL(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.cylindrical.RGB_to_HCL` definition unit
    tests methods.
    """

    def test_RGB_to_HCL(self):
        """Test :func:`colour.models.rgb.cylindrical.RGB_to_HCL` definition."""

        np.testing.assert_almost_equal(
            RGB_to_HCL(np.array([0.45620519, 0.03081071, 0.04091952])),
            np.array([-0.03167854, 0.2841715, 0.22859647]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_HCL(np.array([1.00000000, 2.00000000, 0.50000000])),
            np.array([1.83120102, 1.0075282, 1.00941024]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_HCL(np.array([2.00000000, 1.00000000, 0.50000000])),
            np.array([0.30909841, 1.0075282, 1.00941024]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            RGB_to_HCL(np.array([0.50000000, 1.00000000, 2.00000000])),
            np.array([-2.40349351, 1.0075282, 1.00941024]),
            decimal=7,
        )

    def test_n_dimensional_RGB_to_HCL(self):
        """
        Test :func:`colour.models.rgb.cylindrical.RGB_to_HCL` definition
        n-dimensional arrays support.
        """

        RGB = np.array([0.45620519, 0.03081071, 0.04091952])
        HCL = RGB_to_HCL(RGB)

        RGB = np.tile(RGB, (6, 1))
        HCL = np.tile(HCL, (6, 1))
        np.testing.assert_almost_equal(RGB_to_HCL(RGB), HCL, decimal=7)

        RGB = np.reshape(RGB, (2, 3, 3))
        HCL = np.reshape(HCL, (2, 3, 3))
        np.testing.assert_almost_equal(RGB_to_HCL(RGB), HCL, decimal=7)

    def test_domain_range_scale_RGB_to_HCL(self):
        """
        Test :func:`colour.models.rgb.cylindrical.RGB_to_HCL` definition
        domain and range scale support.
        """

        RGB = np.array([0.45620519, 0.03081071, 0.04091952])
        HCL = RGB_to_HCL(RGB)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    RGB_to_HCL(RGB * factor), HCL * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_RGB_to_HCL(self):
        """
        Test :func:`colour.models.rgb.cylindrical.RGB_to_HCL` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            RGB = np.array(case)
            RGB_to_HCL(RGB)


class TestHCL_to_RGB(unittest.TestCase):
    """
    Define :func:`colour.models.rgb.cylindrical.HCL_to_RGB` definition unit
    tests methods.
    """

    def test_HCL_to_RGB(self):
        """Test :func:`colour.models.rgb.cylindrical.HCL_to_RGB` definition."""

        np.testing.assert_almost_equal(
            HCL_to_RGB(np.array([-0.03167854, 0.28417150, 0.22859647])),
            np.array([0.45620333, 0.03081048, 0.04091925]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            HCL_to_RGB(np.array([1.00000000, 2.00000000, 0.50000000])),
            np.array([0.92186029, 0.71091922, -2.26364935]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            HCL_to_RGB(np.array([2.00000000, 1.00000000, 0.50000000])),
            np.array([-0.31368585, 1.00732462, -0.51534497]),
            decimal=7,
        )

        np.testing.assert_almost_equal(
            HCL_to_RGB(np.array([0.50000000, 1.00000000, 2.00000000])),
            np.array([3.88095422, 3.11881934, 2.40881719]),
            decimal=7,
        )

    def test_n_dimensional_HCL_to_RGB(self):
        """
        Test :func:`colour.models.rgb.cylindrical.HCL_to_RGB` definition
        n-dimensional arrays support.
        """

        HCL = np.array([0.99603944, 0.87347144, 0.24350795])
        RGB = HCL_to_RGB(HCL)

        HCL = np.tile(HCL, (6, 1))
        RGB = np.tile(RGB, (6, 1))
        np.testing.assert_almost_equal(HCL_to_RGB(HCL), RGB, decimal=7)

        HCL = np.reshape(HCL, (2, 3, 3))
        RGB = np.reshape(RGB, (2, 3, 3))
        np.testing.assert_almost_equal(HCL_to_RGB(HCL), RGB, decimal=7)

    def test_domain_range_scale_HCL_to_RGB(self):
        """
        Test :func:`colour.models.rgb.cylindrical.HCL_to_RGB` definition
        domain and range scale support.
        """

        HCL = np.array([0.99603944, 0.87347144, 0.24350795])
        RGB = HCL_to_RGB(HCL)

        d_r = (("reference", 1), ("1", 1), ("100", 100))
        for scale, factor in d_r:
            with domain_range_scale(scale):
                np.testing.assert_almost_equal(
                    HCL_to_RGB(HCL * factor), RGB * factor, decimal=7
                )

    @ignore_numpy_errors
    def test_nan_HCL_to_RGB(self):
        """
        Test :func:`colour.models.rgb.cylindrical.HCL_to_RGB` definition nan
        support.
        """

        cases = [-1.0, 0.0, 1.0, -np.inf, np.inf, np.nan]
        cases = set(permutations(cases * 3, r=3))
        for case in cases:
            HCL = np.array(case)
            HCL_to_RGB(HCL)


if __name__ == "__main__":
    unittest.main()
